"""Functions to create stimuli."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from emodelrunner.stimuli import Pulse


# adapted from bglibpy.cell
def add_pulse(stimulus, soma_loc):
    """Return Pulse Stimulus.

    Args:
        stimulus (dict): contains the stimulus configuration data
        soma_loc (bluepyopt.ephys.locations.NrnSeclistCompLocation): location of the soma

    Returns:
        Pulse: pulse stimulus
    """
    return Pulse(
        location=soma_loc,
        delay=float(stimulus["Delay"]),
        duration=float(stimulus["Duration"]),
        amp=float(stimulus["AmpStart"]),
        frequency=float(stimulus["Frequency"]),
        width=float(stimulus["Width"]),
    )


def load_pulses(soma_loc, stim_path="protocols/stimuli.json"):
    """Return a list of pulse stimuli.

    Args:
        soma_loc (bluepyopt.ephys.locations.NrnSeclistCompLocation):
            location of the soma
        stim_path (str): path to the pulse stimuli file

    Raises:
        NotImplementedError: if stim["Pattern"] is not "Pulse" in simuli file

    Returns:
        list of Pulse stimuli
    """
    pulse_stims = []
    with open(stim_path, "r", encoding="utf-8") as f:
        stimuli = json.load(f)

    for _, stim in stimuli.items():
        if "Pattern" in stim and stim["Pattern"] == "Pulse":
            pulse_stims.append(add_pulse(stim, soma_loc))
        else:
            raise NotImplementedError(
                "Stimulus other than Pulse in stimuli file. Not implemented yet."
            )
    return pulse_stims



"""Creates .hoc from cell."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import shutil

from emodelrunner.load import (
    load_config,
    get_release_params,
    get_rin_exp_voltage_base,
)
from emodelrunner.create_cells import create_cell_using_config
from emodelrunner.create_hoc_tools import (
    create_synapse_hoc,
    create_simul_hoc,
    create_run_hoc,
    create_main_protocol_hoc,
)
from emodelrunner.parsing_utilities import get_parser_args, set_verbosity


def write_hoc(hoc_dir, hoc_file_name, hoc):
    """Write hoc file.

    Args:
        hoc_dir (str): directory to write the file in
        hoc_file_name (str): name to give to the file
        hoc (str): content of the file
    """
    hoc_path = os.path.join(hoc_dir, hoc_file_name)
    with open(hoc_path, "w", encoding="utf-8") as hoc_file:
        hoc_file.write(hoc)


def write_hocs(
    hoc_paths, cell_hoc, simul_hoc, run_hoc, syn_hoc=None, main_protocol_hoc=None
):
    """Write hoc files.

    Args:
        hoc_paths (HocPaths): contains paths of the hoc files to be created
        cell_hoc (str): content of the cell hoc file
        simul_hoc (str): content of the 'create simulation' hoc file
        run_hoc (str): content of the hoc file meant to run the simulation
        syn_hoc (str): content of the synapses hoc file
        main_protocol_hoc (str): content of the main protocol hoc file
    """
    # cell hoc
    write_hoc(hoc_paths.hoc_dir, hoc_paths.cell_hoc_filename, cell_hoc)

    # createsimulation.hoc
    write_hoc(hoc_paths.hoc_dir, hoc_paths.simul_hoc_filename, simul_hoc)

    # run.hoc
    write_hoc(hoc_paths.hoc_dir, hoc_paths.run_hoc_filename, run_hoc)

    # synapses hoc
    if syn_hoc is not None:
        write_hoc(hoc_paths.syn_dir, hoc_paths.syn_hoc_filename, syn_hoc)

    # main protocol hoc
    if main_protocol_hoc is not None:
        write_hoc(
            hoc_paths.hoc_dir, hoc_paths.main_protocol_filename, main_protocol_hoc
        )


def get_hoc(config):
    """Return the hoc scripts as strings.

    Args:
        config (configparser.ConfigParser): configuration

    Raises:
        KeyError: If no voltage_base feature is found for Rin in feature file

    Returns:
        (str, str, str, str): cell_hoc, syn_hoc, simul_hoc, run_hoc

        the hoc scripts for the cell, the synapses, the simulation and the run_launcher.
        syn_hoc is None if the synapses were not added
    """
    # pylint: disable=too-many-locals
    # get directories and filenames from config
    cell_template_path = config.get("Paths", "cell_template_path")
    run_hoc_template_path = config.get("Paths", "run_hoc_template_path")
    createsimulation_template_path = config.get(
        "Paths", "createsimulation_template_path"
    )
    synapses_template_path = config.get("Paths", "synapses_template_path")
    main_protocol_template_path = config.get("Paths", "main_protocol_template_path")
    add_synapses = config.getboolean("Synapses", "add_synapses")
    syn_temp_name = config.get("Synapses", "hoc_synapse_template_name")
    hoc_paths = config.hoc_paths_args()
    apical_point_isec = config.get("Protocol", "apical_point_isec")

    constants_args = {
        "emodel": config.get("Cell", "emodel"),
        "morph_path": config.get("Paths", "morph_path"),
        "gid": config.getint("Cell", "gid"),
        "dt": config.getfloat("Sim", "dt"),
        "celsius": config.getfloat("Cell", "celsius"),
        "v_init": config.getfloat("Cell", "v_init"),
        "mtype": config.get("Morphology", "mtype"),
    }

    # get the protocols definitions
    protocols_filename = config.get("Paths", "prot_path")
    with open(protocols_filename, "r", encoding="utf-8") as protocol_file:
        protocol_definitions = json.load(protocol_file)
    if "__comment" in protocol_definitions:
        del protocol_definitions["__comment"]

    # handle MainProtocol case
    if "Main" in protocol_definitions.keys():
        main_protocol = True
        new_prot_defs = {}
        for prot in protocol_definitions["Main"]["pre_protocols"]:
            new_prot_defs[prot] = protocol_definitions[prot]
        for prot in protocol_definitions["Main"]["other_protocols"]:
            new_prot_defs[prot] = protocol_definitions[prot]
    else:
        main_protocol = False
        new_prot_defs = protocol_definitions

    # get cell
    cell = create_cell_using_config(config)
    release_params = get_release_params(config)

    # get cell hoc
    cell_hoc = cell.create_custom_hoc(
        release_params,
        template_path=cell_template_path,
        syn_dir=hoc_paths.syn_dir_for_hoc,
        syn_hoc_filename=hoc_paths.syn_hoc_filename,
        syn_temp_name=syn_temp_name,
    )

    simul_hoc = create_simul_hoc(
        template_path=createsimulation_template_path,
        add_synapses=add_synapses,
        hoc_paths=hoc_paths,
        constants_args=constants_args,
        protocol_definitions=new_prot_defs,
        apical_point_isec=apical_point_isec,
    )

    run_hoc = create_run_hoc(
        template_path=run_hoc_template_path,
        main_protocol=main_protocol,
    )

    if main_protocol:
        rin_exp_voltage_base = get_rin_exp_voltage_base(
            config.get("Paths", "features_path")
        )

        main_protocol_hoc = create_main_protocol_hoc(
            template_path=main_protocol_template_path,
            protocol_definitions=protocol_definitions,
            rin_exp_voltage_base=rin_exp_voltage_base,
            mtype=constants_args["mtype"],
        )
    else:
        main_protocol_hoc = None

    # get synapse hoc
    if cell.add_synapses:
        # load synapse config data
        syn_mech_args = config.syn_mech_args()

        # get synapse hoc
        syn_hoc = create_synapse_hoc(
            syn_mech_args=syn_mech_args,
            syn_hoc_dir=hoc_paths.syn_dir_for_hoc,
            template_path=synapses_template_path,
            gid=cell.gid,
            dt=constants_args["dt"],
            synapses_template_name=syn_temp_name,
        )
    else:
        syn_hoc = None

    return cell_hoc, syn_hoc, simul_hoc, run_hoc, main_protocol_hoc


def copy_features_hoc(config):
    """Copy features hoc file into cell directory."""
    features_original_path = config.get("Paths", "features_hoc_template_path")
    features_new_path = os.path.join(
        config.get("Paths", "memodel_dir"), config.get("Paths", "features_hoc_file")
    )
    shutil.copy(features_original_path, features_new_path)


if __name__ == "__main__":
    args = get_parser_args()
    set_verbosity(args.verbosity)

    config_ = load_config(config_path=args.config_path)

    cell_hoc_, syn_hoc_, simul_hoc_, run_hoc_, main_protocol_hoc_ = get_hoc(
        config=config_
    )

    hoc_paths_ = config_.hoc_paths_args()
    if main_protocol_hoc_:
        copy_features_hoc(config_)
    write_hocs(
        hoc_paths_,
        cell_hoc_,
        simul_hoc_,
        run_hoc_,
        syn_hoc_,
        main_protocol_hoc_,
    )



"""Frame Classes for the GUI."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=wrong-import-position, too-many-ancestors, too-many-lines, import-error
import tkinter as tk
from tkinter import ttk
import numpy as np
import matplotlib

matplotlib.use("TkAgg")
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure

try:
    from matplotlib.backends.backend_tkagg import NavigationToolbar2TkAgg
except ImportError:
    from matplotlib.backends.backend_tkagg import (
        NavigationToolbar2Tk as NavigationToolbar2TkAgg,
    )

from emodelrunner.GUI_utils.plotshape import get_morph_lines
from emodelrunner.GUI_utils.style import style_dict


def positive_int_callback(input_):
    """Accepts only digits or '' as entry.

    Args:
        input_ (str): input to check

    Returns:
        bool: True if input is an int or an empty string, False otherwise
    """
    return input_.isdigit() or input_ == ""


def float_callback(input_):
    """Accepts only a float or '' as entry.

    Args:
        input_ (str): input to check

    Returns:
        bool: True if input is a float or an empty string, False otherwise
    """
    if input_ != "":
        try:
            float(input_)
        except (ValueError, TypeError):
            return False
    return True


class ToolbarCustom(NavigationToolbar2TkAgg):
    """Matplotlib toolbar class."""

    def set_message(self, msg):
        """Do not show message. This is unstable.

        Args:
            msg (str): message
        """


class FrameSetIntFromEntry(ttk.Frame):
    """Frame containing an entry for input."""

    def __init__(self, parent, gui, attr_name, label):
        """Constructor.

        Args:
            parent (ttk.Frame): parent frame in which to embed this frame
            gui (GUI): main class containing main frames and simulation
            attr_name (str): attribute of gui.simulation that can be changed with the entry
            label (str): text describing the attribute to display
        """
        ttk.Frame.__init__(self, parent, style="TFrame")

        # label
        self.label = ttk.Label(self, text=f"{label}:")

        # to enforce only digits in entry
        self.reg = self.register(positive_int_callback)

        # string variable to be binded to entry
        self.sv = tk.StringVar()
        self.sv.set(str(getattr(gui.simulation, attr_name)))  # default value
        self.sv.trace_add(
            "write",
            lambda name, index, mode, sv=self.sv: self.get_value(gui, attr_name),
        )

        # entry
        self.entry = ttk.Entry(
            self,
            textvariable=self.sv,
            font=style_dict["base_font"],
            width=style_dict["entry_width"],
        )
        self.entry.config(validate="key", validatecommand=(self.reg, "%P"))

        # grid
        self.label.grid(row=0, column=0, sticky=tk.W)
        self.entry.grid(row=0, column=1, sticky=tk.E)

        self.columnconfigure(1, weight=1)  # only center column grows

    def get_value(self, gui, attr_name):
        """Put input value in simulation attribute.

        Args:
            gui (GUI): main class containing main frames and simulation
            attr_name (str): attribute of gui.simulation that can be changed with the entry
        """
        value = self.entry.get()
        if value == "":
            value = 0
        try:
            setattr(gui.simulation, attr_name, int(value))
            gui.config_has_changed()
        except (ValueError, TypeError):
            tk.messagebox.showerror(
                f"Bad {attr_name} value",
                "Must be an int.",
            )

    def disable(self):
        """Set entry state to disabled."""
        self.entry.state(["disabled"])

    def enable(self):
        """Set entry state to not disabled."""
        self.entry.state(["!disabled"])


class FrameStepStimulus(ttk.Frame):
    """Frame containing step stimulus value input."""

    def __init__(self, parent, gui):
        """Constructor.

        Args:
            parent (ttk.Frame): parent frame in which to embed this frame
            gui (GUI): main class containing main frames and simulation
        """
        ttk.Frame.__init__(self, parent, style="TFrame")

        self.step_stim = tk.DoubleVar()
        self.step_stim.set(gui.simulation.step_stim)  # step1 is selected

        self.step_stim_title = ttk.Label(self, text="Select a Step Stimulus")

        self.steps = []
        for step_amp in gui.simulation.steps:
            protocol_step = ttk.Radiobutton(
                self,
                text=f"{step_amp:.3g} nA",
                variable=self.step_stim,
                value=step_amp,
                command=lambda: self.get_step_stim(gui),
            )
            self.steps.append(protocol_step)

        self.custom_step = ttk.Radiobutton(
            self,
            text="Custom stimulus [nA]: ",
            variable=self.step_stim,
            value=gui.simulation.step_stim,
            command=lambda: self.get_custom_step_stim(gui),
        )

        # to enforce only float in entry
        self.reg = self.register(float_callback)

        # string variable to be binded to entry
        self.sv = tk.StringVar()
        self.sv.set("0.0")  # default value
        self.sv.trace_add(
            "write",
            lambda name, index, mode, sv=self.sv: self.get_custom_step_stim(gui),
        )

        # entry
        self.custom_entry = ttk.Entry(
            self,
            textvariable=self.sv,
            font=style_dict["base_font"],
            width=style_dict["entry_width"],
        )
        self.custom_entry.config(validate="key", validatecommand=(self.reg, "%P"))
        self.custom_entry.state(["!disabled"])

        self.step_stim_title.grid(row=0, column=0, sticky=tk.W)
        self.custom_step.grid(row=1, column=0, sticky=tk.W)
        self.custom_entry.grid(row=1, column=1, sticky=tk.E)
        step_row = 1
        for protocol_step in self.steps:
            step_row += 1
            protocol_step.grid(row=step_row, column=0, sticky=tk.W)

        self.columnconfigure(1, weight=1)  # only center column grows
        for i in range(step_row + 1):
            self.rowconfigure(i, weight=1)

    def get_step_stim(self, gui):
        """Put selected step stim value into simulation attribute.

        Args:
            gui (GUI): main class containing main frames and simulation
        """
        self.custom_entry.state(["disabled"])

        value = self.step_stim.get()
        gui.simulation.step_stim = value
        gui.config_has_changed()

    def get_custom_step_stim(self, gui):
        """Enable input and put input value into simulation attribute.

        Args:
            gui (GUI): main class containing main frames and simulation
        """
        self.custom_entry.state(["!disabled"])

        value = self.custom_entry.get()
        if value == "":
            value = 0
        try:
            gui.simulation.step_stim = float(value)
            gui.config_has_changed()
        except (ValueError, TypeError):
            tk.messagebox.showerror(
                "Bad step stimulus value",
                "Must be a float.",
            )


class FrameHoldStimulus(ttk.Frame):
    """Frame containing holding stimulus value input."""

    def __init__(self, parent, gui):
        """Constructor.

        Args:
            parent (ttk.Frame): parent frame in which to embed this frame
            gui (GUI): main class containing main frames and simulation
        """
        ttk.Frame.__init__(self, parent, style="TFrame")

        self.hold_stim = tk.DoubleVar()
        self.hold_stim.set(gui.simulation.hypamp)  # default hold stim is selected

        self.hold_stim_title = ttk.Label(self, text="Select a Holding Stimulus")

        self.hypamps = []
        for amp in gui.simulation.hypamps:
            protocol_hypamp = ttk.Radiobutton(
                self,
                text=f"{amp:.3g} nA",
                variable=self.hold_stim,
                value=amp,
                command=lambda: self.get_hold_stim(gui),
            )
            self.hypamps.append(protocol_hypamp)

        self.custom_hold = ttk.Radiobutton(
            self,
            text="Custom stimulus [nA]: ",
            variable=self.hold_stim,
            value=gui.simulation.hypamp,
            command=lambda: self.get_custom_hold_stim(gui),
        )

        # to enforce only float in entry
        self.reg = self.register(float_callback)

        # string variable to be binded to entry
        self.sv = tk.StringVar()
        self.sv.set("0.0")  # default value
        self.sv.trace_add(
            "write",
            lambda name, index, mode, sv=self.sv: self.get_custom_hold_stim(gui),
        )

        # entry
        self.custom_entry = ttk.Entry(
            self,
            textvariable=self.sv,
            font=style_dict["base_font"],
            width=style_dict["entry_width"],
        )
        self.custom_entry.config(validate="key", validatecommand=(self.reg, "%P"))
        self.custom_entry.state(["!disabled"])

        self.hold_stim_title.grid(row=0, column=0, sticky=tk.W)
        self.custom_hold.grid(row=1, column=0, sticky=tk.W)
        self.custom_entry.grid(row=1, column=1, sticky=tk.E)
        hold_row = 1
        for protocol_hypamp in self.hypamps:
            hold_row += 1
            protocol_hypamp.grid(row=hold_row, column=0, sticky=tk.W)

        self.columnconfigure(1, weight=1)  # only center column grows
        for i in range(hold_row + 1):
            self.rowconfigure(i, weight=1)

    def get_hold_stim(self, gui):
        """Put selected holding stim value into simulation attribute.

        Args:
            gui (GUI): main class containing main frames and simulation
        """
        self.custom_entry.state(["disabled"])

        value = self.hold_stim.get()
        gui.simulation.hypamp = value
        gui.config_has_changed()

    def get_custom_hold_stim(self, gui):
        """Enable input and put input value into simulation attribute.

        Args:
            gui (GUI): main class containing main frames and simulation
        """
        self.custom_entry.state(["!disabled"])

        value = self.custom_entry.get()
        if value == "":
            value = 0
        try:
            gui.simulation.hypamp = float(value)
            gui.config_has_changed()
        except (ValueError, TypeError):
            tk.messagebox.showerror(
                "Bad holding stimulus value",
                "Must be a float.",
            )


class FrameStepProtocol(ttk.Frame):
    """Frame containing step stimulus-related input."""

    def __init__(self, parent, gui):
        """Constructor.

        Args:
            parent (ttk.Frame): parent frame in which to embed this frame
            gui (GUI): main class containing main frames and simulation
        """
        ttk.Frame.__init__(self, parent, style="TFrame")

        # step stimulus
        self.frame_step_stim = FrameStepStimulus(self, gui)
        self.frame_step_delay = FrameSetIntFromEntry(
            self, gui, "step_delay", "Step stimulus delay [ms]"
        )
        self.frame_step_duration = FrameSetIntFromEntry(
            self, gui, "step_duration", "Step stimulus duration [ms]"
        )

        # holding stimulus
        self.frame_hold_stim = FrameHoldStimulus(self, gui)
        self.frame_hold_step_delay = FrameSetIntFromEntry(
            self, gui, "hold_step_delay", "Holding stimulus delay [ms]"
        )
        self.frame_hold_step_duration = FrameSetIntFromEntry(
            self,
            gui,
            "hold_step_duration",
            "Holding stimulus duration [ms]",
        )

        # display on grid
        self.frame_step_stim.grid(row=0, column=0, sticky=(tk.W, tk.E))
        self.frame_step_delay.grid(row=1, column=0, sticky=(tk.W, tk.E))
        self.frame_step_duration.grid(row=2, column=0, sticky=(tk.W, tk.E))

        self.frame_hold_stim.grid(row=3, column=0, sticky=(tk.W, tk.E))
        self.frame_hold_step_delay.grid(row=4, column=0, sticky=(tk.W, tk.E))
        self.frame_hold_step_duration.grid(row=5, column=0, sticky=(tk.W, tk.E))

        self.columnconfigure(0, weight=1)
        for i in range(6):
            self.rowconfigure(i, weight=1)


class FrameProtocols(ttk.LabelFrame):
    """Frame containing protocol-related inputs."""

    def __init__(self, parent, gui, title):
        """Constructor.

        Args:
            parent (ttk.Frame): parent frame in which to embed this frame
            gui (GUI): main class containing main frames and simulation
            title (ttk.Label): frame title to display
        """
        ttk.LabelFrame.__init__(self, parent, style="Boxed.TFrame", labelwidget=title)

        self.frame_sim_duration = FrameSetIntFromEntry(
            self, gui, "total_duration", "Simulation time [ms]"
        )
        self.frame_step_protocol = FrameStepProtocol(self, gui)

        self.frame_sim_duration.grid(row=0, column=0, sticky=(tk.W, tk.E))
        self.frame_step_protocol.grid(row=1, column=0, sticky=(tk.W, tk.E, tk.N, tk.S))

        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, weight=12)


class FrameConfig(ttk.Frame):
    """Frame containing all inputs."""

    def __init__(self, parent, gui):
        """Constructor.

        Args:
            parent (ttk.Frame): parent frame in which to embed this frame
            gui (GUI): main class containing main frames and simulation
        """
        ttk.Frame.__init__(self, parent, style="TFrame")

        title_config_fig = ttk.Label(self, text="Display configuration")
        self.frame_config_fig = FrameConfigFig(self, gui, title_config_fig)

        title_protocols = ttk.Label(
            self, text="Simulation & Step stimulus configuration"
        )
        self.frame_protocols = FrameProtocols(self, gui, title_protocols)

        self.frame_config_fig.grid(row=0, column=0, sticky=(tk.W, tk.E, tk.N, tk.S))
        self.frame_protocols.grid(row=1, column=0, sticky=(tk.W, tk.E, tk.N, tk.S))

        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, weight=10)


class FrameButtons(ttk.Frame):
    """Frame containing buttons to (re-)start and pause simulation."""

    def __init__(self, parent, gui):
        """Constructor.

        Args:
            parent (ttk.Frame): parent frame in which to embed this frame
            gui (GUI): main class containing main frames and simulation
        """
        ttk.Frame.__init__(self, parent, style="TFrame")

        self.start_button = ttk.Button(
            self, text="Start", command=gui.start, style="ControlSimul.TButton"
        )

        self.pause_button = ttk.Button(
            self,
            text="Pause",
            command=gui.pause,
            state=tk.DISABLED,
            style="ControlSimul.TButton",
        )

        self.continue_button = ttk.Button(
            self,
            text="Continue",
            command=gui.continue_simul,
            state=tk.DISABLED,
            style="ControlSimul.TButton",
        )

        self.start_button.grid(row=0, column=0)
        self.pause_button.grid(row=0, column=1)
        self.continue_button.grid(row=0, column=2)

    def simul_running(self):
        """Disable continue button, enable pause button."""
        self.pause_button["state"] = tk.NORMAL
        self.continue_button["state"] = tk.DISABLED

    def simul_on_pause(self):
        """Disable pause button, enable continue button."""
        self.pause_button["state"] = tk.DISABLED
        self.continue_button["state"] = tk.NORMAL

    def simul_ended(self):
        """Disable pause & continue buttons."""
        self.pause_button["state"] = tk.DISABLED
        self.continue_button["state"] = tk.DISABLED


class FrameFigures(ttk.Frame):
    """Frame containing the morphology and the voltage figures."""

    def __init__(
        self,
        parent,
        simulation,
        plot_3d=False,
        toolbar_on=False,
        figsize="medium",
        val_min=-80,
        val_max=30,
    ):
        """Constructor.

        Args:
            parent (ttk.Frame): parent frame in which to embed this frame
            simulation (NeuronSimulation): contains simulation (and cell) data
            plot_3d (bool): set to True to plot the cell shapes in 3D
            toolbar_on (bool): set to True to display the matplotlib toolbars
            figsize (str): figures size. can be "small", "medium", or "large".
            val_min (int): minimum voltage for colormap
            val_max (int): maximum voltage for colormap
        """
        ttk.Frame.__init__(self, parent, style="TFrame")

        self.xaxis = 2  # z
        self.yaxis = 0  # x
        self.zaxis = 1  # y
        self.plot_3d = plot_3d
        self.figsize = figsize

        self.val_min = val_min
        self.val_max = val_max

        # ---
        # figure for neuron visualisation
        # ---
        fig_morph = Figure()
        if self.plot_3d:
            self.ax_morph = fig_morph.add_subplot(111, projection="3d")
        else:
            self.ax_morph = fig_morph.add_subplot(111)
            self.ax_morph.set_aspect(aspect=1)

        # get data and axis lims for left morph figure
        _, self.old_vals, _ = get_morph_lines(
            ax=self.ax_morph,
            sim=simulation.sim,
            val_min=self.val_min,
            val_max=self.val_max,
            do_plot=True,
            plot_3d=self.plot_3d,
            xaxis=self.xaxis,
            yaxis=self.yaxis,
            zaxis=self.zaxis,
        )
        self.vals_last_draw = self.old_vals.copy()

        # resize & adjust figure for not cutting axis
        self.set_fig_morph_display(fig_morph)

        # create canva
        self.canva_morph = FigureCanvasTkAgg(fig_morph, self)
        self.canva_morph.get_tk_widget().grid(
            row=0, column=0, sticky=(tk.W, tk.E, tk.N, tk.S)
        )

        # for interactive 3d rotating with mouse
        if self.plot_3d:
            self.get_interactive_3d_rotation(self.canva_morph, self.ax_morph)

        # ---
        # figure for neuron visualisation with synapses
        # ---
        fig_morph_syn = Figure()
        if self.plot_3d:
            self.ax_morph_syn = fig_morph_syn.add_subplot(111, projection="3d")
        else:
            self.ax_morph_syn = fig_morph_syn.add_subplot(111)
            self.ax_morph_syn.set_aspect(aspect=1)

        # get data and axis lims
        get_morph_lines(
            ax=self.ax_morph_syn,
            sim=simulation.sim,
            do_plot=True,
            cmap=None,
            plot_3d=self.plot_3d,
            xaxis=self.xaxis,
            yaxis=self.yaxis,
            zaxis=self.zaxis,
        )

        # resize & adjust figure for not cutting axis
        self.set_fig_morph_display(fig_morph_syn)

        # create canva
        self.canva_morph_syn = FigureCanvasTkAgg(fig_morph_syn, self)
        # save background for blit
        self.ax_morph_syn_bg = self.canva_morph_syn.copy_from_bbox(
            self.ax_morph_syn.bbox
        )
        self.canva_morph_syn.get_tk_widget().grid(
            row=0, column=1, sticky=(tk.W, tk.E, tk.N, tk.S)
        )

        # for interactive 3d rotating with mouse
        if self.plot_3d:
            self.get_interactive_3d_rotation(self.canva_morph_syn, self.ax_morph_syn)

        # ---
        # figure for voltage evolution
        # ---
        fig_volt = Figure()
        self.ax_volt = fig_volt.add_subplot(111)
        self.set_axis(x_max=simulation.protocol.total_duration)

        # set fig size
        self.set_fig_volt_display(fig_volt)

        # create canva
        self.canva_volt = FigureCanvasTkAgg(fig_volt, self)
        self.canva_volt.get_tk_widget().grid(
            row=2, column=0, columnspan=2, sticky=(tk.W, tk.E, tk.N, tk.S)
        )

        # add matplotlib toolbar
        if toolbar_on:
            self.set_toolbars()

        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)
        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, weight=1)
        self.rowconfigure(2, weight=1)
        self.rowconfigure(3, weight=1)

    def set_fig_morph_display(self, fig):
        """Set shape figure size and adjustment.

        Args:
            fig (matplotlib.figure.Figure): figure to adjust
        """
        if self.figsize == "small":
            fig.set_size_inches((3, 3))
        elif self.figsize == "large":
            fig.set_size_inches((5, 5))
        else:
            fig.set_size_inches((4, 4))
        if self.plot_3d:
            if self.figsize:
                fig.subplots_adjust(right=0.8, left=0.2)
            else:
                fig.subplots_adjust(right=0.85)
        else:
            fig.subplots_adjust(right=0.98, top=0.98, bottom=0.15, left=0.20)

    def set_fig_volt_display(self, fig):
        """Set shape figure size and adjustment.

        Args:
            fig (matplotlib.figure.Figure): figure to adjust
        """
        if self.figsize == "small":
            fig.set_size_inches((4.5, 2))
        elif self.figsize == "large":
            fig.set_size_inches((7.5, 3))
        else:
            fig.set_size_inches((6, 2.5))
        fig.subplots_adjust(bottom=0.2)  # to avoid the xlabel being cut

    def set_toolbars(self):
        """Set a matplotlib toolbar for each figure."""
        # self.ax_morph.format_coord = lambda x, y: ""
        toolbar_frame_morph = tk.Frame(self)
        toolbar_frame_morph.grid(row=1, column=0, sticky=tk.E)
        ToolbarCustom(self.canva_morph, toolbar_frame_morph)

        # self.ax_morph_syn.format_coord = lambda x, y: ""
        toolbar_frame_morph_syn = tk.Frame(self)
        toolbar_frame_morph_syn.grid(row=1, column=1, sticky=tk.E)
        ToolbarCustom(self.canva_morph_syn, toolbar_frame_morph_syn)

        # self.ax_volt.format_coord = lambda x, y: ""
        toolbar_frame_volt = tk.Frame(self)
        toolbar_frame_volt.grid(row=3, column=0, columnspan=2)
        ToolbarCustom(self.canva_volt, toolbar_frame_volt)

    @staticmethod
    def get_interactive_3d_rotation(canva, ax):
        """Connect events to canva to enable rotative 3d plots with mouse.

        Args:
            canva (matplotlib.backends.backend_tkagg.FigureCanvasTkAgg): canva
            ax (matplotlib.axes.Axes): axes
        """
        # pylint: disable=protected-access
        canva.mpl_connect("button_press_event", ax._button_press)
        canva.mpl_connect("button_release_event", ax._button_release)
        canva.mpl_connect("motion_notify_event", ax._on_move)

    def set_axis(self, x_min=0, x_max=3000, y_min=-90, y_max=40):
        """Set the voltage figure's axis.

        Args:
            x_min (float): min value on x axis
            x_max (float): max value on x axis
            y_min (float): min value on y axis
            y_max (float): max value on y axis
        """
        self.ax_volt.set_xlim([x_min, x_max])
        self.ax_volt.set_ylim([y_min, y_max])
        self.ax_volt.set_xlabel("t [ms]")
        self.ax_volt.set_ylabel("v [mV]")

    def check_change(self, root, simulation):
        """Checks the voltage change in the cell sections.

        Update display if change is significant.

        Args:
            root (tk.Tk): root of the GUI
            simulation (NeuronSimulation): contains simulation (and cell) data

        Returns:
            bool: True if the cell morphology with color-coded voltage figure has been updated
        """
        # here, update is True if any of the morph lines has a significant change of voltage.
        morph_lines, old_vals, update = get_morph_lines(
            ax=self.ax_morph,
            sim=simulation.sim,
            val_min=self.val_min,
            val_max=self.val_max,
            do_plot=False,
            old_vals=self.old_vals,
            vals_last_draw=self.vals_last_draw,
        )
        if update:
            self.old_vals = old_vals
            self.display(root, simulation, morph_lines)
            return True
        return False

    def display(self, root, simulation, morph_lines=None):
        """Update both figures display.

        Args:
            root (tk.Tk): root of the GUI
            simulation (NeuronSimulation): contains simulation (and cell) data
            morph_lines (list of matplotlib.Line2D): list of lines to be actualized
                if they already have been computed, else None
        """
        # get voltage data
        t, v = simulation.get_voltage()

        # update data in Line2D
        if self.ax_volt.lines:
            line = self.ax_volt.lines[0]
            line.set_xdata(t)
            line.set_ydata(v)
        else:
            (line,) = self.ax_volt.plot(t, v)

        # draw voltage plot to canva
        self.ax_volt.draw_artist(line)
        self.canva_volt.blit(self.ax_volt.bbox)

        # do not blit too much on top of figure, or else
        # 'older' lines tend to stack in the background
        # and bias the perceived color of the line.
        # so redraw everything when there is a big change in color
        if morph_lines is not None:
            for line in morph_lines:
                self.ax_morph.draw_artist(line)

            self.vals_last_draw = self.old_vals.copy()
            self.canva_morph.draw()
        else:
            morph_lines, self.old_vals, force_draw = get_morph_lines(
                ax=self.ax_morph,
                sim=simulation.sim,
                val_min=self.val_min,
                val_max=self.val_max,
                do_plot=False,
                old_vals=self.old_vals,
                vals_last_draw=self.vals_last_draw,
            )

            for line in morph_lines:
                self.ax_morph.draw_artist(line)
            if force_draw:
                self.canva_morph.draw()
                self.vals_last_draw = self.old_vals.copy()
            else:
                self.canva_morph.blit(self.ax_morph.bbox)

        # update tkinter display
        root.update()

    def update_syn_display(self, root, simulation, size_scatter=6):
        """Update the display of the synapses on the right figure.

        Args:
            root (tk.Tk): root of the GUI
            simulation (NeuronSimulation): contains simulation (and cell) data
            size_scatter (int): size of synapses for scatter plot
        """
        # get all synapses previously drawn
        drawn_syns = self.ax_morph_syn.collections
        # hide them before redrawing shape
        for syn in drawn_syns:
            syn.set_visible(False)
        self.canva_morph_syn.draw()  # redraw morphology
        # delete the synapses so that the list is not a big bunch of hidden accumulated synapses
        for syn in drawn_syns:
            self.ax_morph_syn.collections.remove(syn)
        # restore background as the morphology without synapses
        self.canva_morph_syn.restore_region(self.ax_morph_syn_bg)

        # create synapse scatterplot
        syn_scatterplot = {}
        for mtype, data in simulation.syn_display_data.items():
            if data:
                data = np.array(data)
                syn_x = data[:, self.xaxis]
                syn_y = data[:, self.yaxis]
                syn_cols = data[:, 3]
                syn_colors = ["red" if x == 1 else "orange" for x in syn_cols]
                if self.plot_3d:
                    syn_z = data[:, self.zaxis]
                    syn_scatterplot[mtype] = self.ax_morph_syn.scatter(
                        xs=syn_x,
                        ys=syn_y,
                        zs=syn_z,
                        s=size_scatter,
                        c=syn_colors,
                        alpha=1,
                    )
                else:
                    syn_scatterplot[mtype] = self.ax_morph_syn.scatter(
                        x=syn_x, y=syn_y, s=size_scatter, c=syn_colors, alpha=1
                    )
            else:
                syn_scatterplot[mtype] = None

        # draw selected synapses
        for mtype in simulation.available_pre_mtypes:
            if mtype in simulation.pre_mtypes and syn_scatterplot[mtype]:
                syn_scatterplot[mtype].set_visible(True)
                self.ax_morph_syn.draw_artist(syn_scatterplot[mtype])
            elif syn_scatterplot[mtype]:
                syn_scatterplot[mtype].set_visible(False)

        # 3d does not support blitting
        if self.plot_3d:
            self.canva_morph_syn.draw()
        # redrawing somehow draws synapse behind shape in 2d.
        else:
            self.canva_morph_syn.blit(self.ax_morph_syn.bbox)

        # update tkinter display
        root.update()

    def restart_volt(self):
        """Clean the voltage figure."""
        (line,) = self.ax_volt.plot([], [])
        self.ax_volt.draw_artist(line)
        self.canva_volt.draw_idle()


class FrameMain(ttk.Frame):
    """Frame containing Figures and launching button."""

    def __init__(self, parent, gui):
        """Constructor.

        Args:
            parent (ttk.Frame): parent frame in which to embed this frame
            gui (GUI): main class containing main frames and simulation
        """
        ttk.Frame.__init__(self, parent, style="TFrame")
        self.frame_buttons = FrameButtons(self, gui)
        self.frame_figures = FrameFigures(
            self, gui.simulation, gui.plot_3d, gui.toolbar_on, gui.figsize
        )

        self.frame_buttons.grid(row=0, column=0)
        self.frame_figures.grid(row=1, column=0, sticky=(tk.W, tk.E, tk.N, tk.S))

    def display(self, root, simulation):
        """Update figures display.

        Args:
            root (tk.Tk): root of the GUI
            simulation (NeuronSimulation): contains simulation (and cell) data
        """
        self.frame_figures.display(root, simulation)

    def check_change(self, root, simulation):
        """Checks the voltage change in the cell sections.

        Update display if change is significant.

        Args:
            root (tk.Tk): root of the GUI
            simulation (NeuronSimulation): contains simulation (and cell) data

        Returns:
            bool: True if the cell morphology with color-coded voltage figure has been updated
        """
        return self.frame_figures.check_change(root, simulation)

    def update_syn_display(self, root, simulation):
        """Update the display of the synapses on the right figure.

        Args:
            Args:
            root (tk.Tk): root of the GUI
            simulation (NeuronSimulation): contains simulation (and cell) data
        """
        self.frame_figures.update_syn_display(root, simulation)

    def restart_volt(self):
        """Clean voltage figure."""
        self.frame_figures.restart_volt()

    def simul_on_pause(self):
        """Disable pause button, enable continue button."""
        self.frame_buttons.simul_on_pause()

    def simul_running(self):
        """Disable continue button, enable pause button."""
        self.frame_buttons.simul_running()

    def simul_ended(self):
        """Disable pause & continue buttons."""
        self.frame_buttons.simul_ended()


class FrameSynapses(ttk.LabelFrame):
    """Frame containing all inputs."""

    def __init__(self, parent, gui, title):
        """Constructor.

        Args:
            parent (ttk.Frame): parent frame in which to embed this frame
            gui (GUI): main class containing main frames and simulation
            title (ttk.Label): frame title to display
        """
        ttk.LabelFrame.__init__(self, parent, style="Boxed.TFrame", labelwidget=title)

        # -- labels --
        self.labels = []
        self.labels.append(ttk.Label(self, text="Synapses"))
        self.labels.append(ttk.Label(self, text="Start time [ms]"))
        self.labels.append(ttk.Label(self, text="Spike Interval [ms]"))
        self.labels.append(ttk.Label(self, text="Spike number"))
        self.labels.append(ttk.Label(self, text="Noise"))

        for i, label in enumerate(self.labels):
            label.grid(row=0, column=i)
            self.columnconfigure(i, weight=1)

        self.rowconfigure(0, weight=1)

        # -- synapses --
        self.reg = self.register(positive_int_callback)

        self.mtype_buttons = []
        self.var_list = []  # 0 or 1 for each button / checkbox values
        self.id_list = []
        self.svs = []  # entries (start, interval, number, noise) values

        # list of lists : start, interval, number and noise
        self.entries = [[] for x in range(4)]

        # -- create buttons & parameter entries --
        for i, id_ in enumerate(gui.simulation.available_pre_mtypes.keys()):
            # pre-cell m-types
            self.var_list.append(tk.IntVar())
            self.var_list[i].set(0)
            self.id_list.append(id_)
            self.mtype_buttons.append(
                ttk.Checkbutton(
                    self,
                    text=gui.simulation.available_pre_mtypes[id_],
                    variable=self.var_list[i],
                    command=lambda: self.load_current_mtype_list(gui),
                    offvalue=0,
                    onvalue=1,
                )
            )
            # synapse start, interval, number, noise
            for j, entry in enumerate(self.entries):
                self.svs.append(tk.StringVar())
                entry.append(
                    ttk.Entry(
                        self,
                        textvariable=self.svs[4 * i + j],
                        font=style_dict["base_font"],
                        width=style_dict["entry_width"],
                    )
                )
                entry[i].config(validate="key", validatecommand=(self.reg, "%P"))
                entry[i].state(["disabled"])

            # set string variables
            self.set_svs(gui, i)

        # -- add buttons & entries on the grid --
        for i, (b, e1, e2, e3, e4) in enumerate(
            zip(*[self.mtype_buttons] + self.entries)
        ):
            b.grid(row=i + 1, column=0, sticky=(tk.W, tk.E), padx=2)
            e1.grid(row=i + 1, column=1, sticky=(tk.E), padx=2)
            e2.grid(row=i + 1, column=2, sticky=(tk.E), padx=2)
            e3.grid(row=i + 1, column=3, sticky=(tk.E), padx=2)
            e4.grid(row=i + 1, column=4, sticky=(tk.E), padx=2)
            self.rowconfigure(i + 1, weight=1)

    def toggle_button(self):
        """Enable/disable entries depending on the button status."""
        for i, var in enumerate(self.var_list):
            if var.get():
                for entry in self.entries:
                    entry[i].state(["!disabled"])
            else:
                for entry in self.entries:
                    entry[i].state(["disabled"])

    def set_svs(self, gui, i):
        """Returns the entry and the variable associated.

        Args:
            gui (GUI): main class containing main frames and simulation
            i (int): mtype index
        """
        # pylint: disable=cell-var-from-loop
        default_var = [
            gui.simulation.syn_start,
            gui.simulation.syn_interval,
            gui.simulation.syn_nmb_of_spikes,
            gui.simulation.syn_noise,
        ]
        for j, var in enumerate(default_var):
            self.svs[4 * i + j].set(str(var))
            self.svs[4 * i + j].trace_add(
                "write",
                lambda name, index, mode, sv=self.svs[
                    4 * i + j
                ]: self.load_current_mtype_list(gui),
            )

    @staticmethod
    def check_variable(x):
        """Returns the variable if it is a positive int. Returns 0 if not.

        Args:
            x (str): variable to check

        Returns:
            int: the variable if it is a positive int
        """
        try:
            if x == "":
                x = 0
            else:
                x = int(x)
                assert x >= 0
        except (ValueError, TypeError, AssertionError):
            tk.messagebox.showerror(
                "Bad parameter value",
                "Must be a positive int.",
            )
            x = 0
        return x

    def load_current_mtype_list(self, gui):
        """Load current mtype list and netstim params.

        Args:
            gui (GUI): main class containing main frames and simulation
        """
        gui.simulation.pre_mtypes = []
        gui.simulation.netstim_params = {}

        for idx, var, e1, e2, e3, e4 in zip(
            *[self.id_list, self.var_list] + self.entries
        ):
            if var.get():
                gui.simulation.pre_mtypes.append(idx)
                v1 = self.check_variable(e1.get())
                v2 = self.check_variable(e2.get())
                v3 = self.check_variable(e3.get())
                v4 = self.check_variable(e4.get())
                params = [v1, v2, v3, v4]
                # synapse netstim param depending on mtype
                # {mtypeidx:[start, interval, number, noise]}
                gui.simulation.netstim_params[idx] = params
        self.toggle_button()
        gui.frames["FrameMain"].update_syn_display(gui.root, gui.simulation)
        gui.config_has_changed()


class FrameConfigFig(ttk.LabelFrame):
    """Frame containing choices for figure display, such as 2d/3d or enabling toolbar."""

    def __init__(self, parent, gui, title):
        """Constructor.

        Args:
            parent (ttk.Frame): parent frame in which to embed this frame
            gui (GUI): main class containing main frames and simulation
            title (ttk.Label): frame title to display
        """
        ttk.LabelFrame.__init__(self, parent, style="Boxed.TFrame", labelwidget=title)

        # 2d / 3d radiobuttons
        self.plot_3d_var = tk.IntVar()
        self.plot_3d_var.set(int(gui.plot_3d))  # default hold stim is selected

        self.plot_2d_button = ttk.Radiobutton(
            self,
            text="2D neuron shape",
            variable=self.plot_3d_var,
            value=0,
            command=lambda: self.load_plot_3d_value(gui),
        )

        self.plot_3d_button = ttk.Radiobutton(
            self,
            text="3D neuron shape",
            variable=self.plot_3d_var,
            value=1,
            command=lambda: self.load_plot_3d_value(gui),
        )

        # toolbar checkbutton
        self.toolbar_var = tk.IntVar()
        self.toolbar_var.set(int(gui.toolbar_on))
        self.toolbar_button = ttk.Checkbutton(
            self,
            text="display figures toolbars",
            variable=self.toolbar_var,
            command=lambda: self.load_toolbar_value(gui),
            offvalue=0,
            onvalue=1,
        )

        # figsize choice
        self.figsize_var = tk.StringVar()
        self.figsize_var.set(str(gui.figsize))
        self.figsize_label = ttk.Label(self, text="Figure size:")
        self.figsize_small_button = ttk.Radiobutton(
            self,
            text="small",
            variable=self.figsize_var,
            value="small",
            command=lambda: self.load_figsize_value(gui),
        )

        self.figsize_medium_button = ttk.Radiobutton(
            self,
            text="medium",
            variable=self.figsize_var,
            value="medium",
            command=lambda: self.load_figsize_value(gui),
        )

        self.figsize_large_button = ttk.Radiobutton(
            self,
            text="large",
            variable=self.figsize_var,
            value="large",
            command=lambda: self.load_figsize_value(gui),
        )

        # put buttons on grid
        self.plot_2d_button.grid(row=0, column=0, columnspan=3, sticky=(tk.W, tk.E))
        self.plot_3d_button.grid(row=1, column=0, columnspan=3, sticky=(tk.W, tk.E))
        self.toolbar_button.grid(row=2, column=0, columnspan=3, sticky=(tk.W, tk.E))
        self.figsize_label.grid(row=3, column=0, columnspan=3, sticky=(tk.W, tk.E))
        self.figsize_small_button.grid(row=4, column=0, sticky=(tk.W, tk.E))
        self.figsize_medium_button.grid(row=4, column=1, sticky=(tk.W, tk.E))
        self.figsize_large_button.grid(row=4, column=2, sticky=(tk.W, tk.E))

        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, weight=1)
        self.rowconfigure(2, weight=1)
        self.rowconfigure(3, weight=1)
        self.rowconfigure(4, weight=1)
        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)
        self.columnconfigure(2, weight=1)

    def load_toolbar_value(self, gui):
        """Change toolbar value in gui and reload figure frame.

        Args:
            gui (GUI): main class containing main frames and simulation
        """
        if self.toolbar_var.get():
            gui.toolbar_on = True
        else:
            gui.toolbar_on = False

        gui.reload_figure_frame()

    def load_plot_3d_value(self, gui):
        """Change figure display in gui and reload figure frame.

        Args:
            gui (GUI): main class containing main frames and simulation
        """
        if self.plot_3d_var.get():
            gui.plot_3d = True
        else:
            gui.plot_3d = False

        gui.reload_figure_frame()

    def load_figsize_value(self, gui):
        """Change figure size in gui and reload figure frame.

        Args:
            gui (GUI): main class containing main frames and simulation
        """
        gui.figsize = self.figsize_var.get()
        gui.reload_figure_frame()



"""Custom Recording class."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from bluepyopt import ephys

logger = logging.getLogger(__name__)


class RecordingCustom(ephys.recordings.CompRecording):
    """Response to stimulus with recording every 0.1 ms.

    Attributes:
        name (str): name of this object
        location (Location): location in the model of the recording
        variable (str): which variable to record from (e.g. 'v')
        varvector (neuron Vector): vector recording the variable
        tvector (neuron Vector): vector recording the time (ms)
        instantiated (bool): whether the object has been instantiated or not

    Args of the parent constructor:

    - name (str): name of this object
    - location (Location): location in the model of the recording
    - variable (str): which variable to record from (e.g. 'v')
    """

    def instantiate(self, sim=None, icell=None):
        """Instantiate recording.

        Args:
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
            icell (neuron cell): cell instantiation in simulator
        """
        logger.debug(
            "Adding compartment recording of %s at %s", self.variable, self.location
        )

        self.varvector = sim.neuron.h.Vector()
        seg = self.location.instantiate(sim=sim, icell=icell)
        self.varvector.record(getattr(seg, f"_ref_{self.variable}"), 0.1)

        self.tvector = sim.neuron.h.Vector()
        self.tvector.record(sim.neuron.h._ref_t, 0.1)  # pylint: disable=W0212

        self.instantiated = True



"""Creates .hoc from cell."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=too-many-arguments
from datetime import datetime
import jinja2

from bluepyopt.ephys.create_hoc import (
    generate_parameters,
    generate_channels_by_location,
    generate_reinitrng,
    range_exprs_to_hoc,
)

from emodelrunner import __version__


class HocStimuliCreator:
    """Class to create the stimuli in hoc.

    Attributes:
        apical_point_isec (int): section index of the apical point
            Set to -1 if there is no apical point
        n_stims (int): total number of protocols to be run by hoc.
            Gets incremented during initiation to enumerate the protocols.
        max_steps (int): A StepProtocol can have multiple steps. This attribute
            counts the maximum steps that the StepProtocol with the most steps has.
            Stays at 0 if there is no StepProtocol.
        reset_step_stimuli (str): hoc script resetting the step stimuli objects
            to be put inhoc file.
        init_step_stimuli (str): hoc script initiating the step stimuli objects
            to be put in hoc file.
        stims_hoc (str): hoc script containing all the protocols to be run by hoc.
            The Protocols supported by this library to be converted to hoc are:

            - StepProtocol
            - StepThresholdProtocol (only inside  Main Protocol)
            - RampProtocol
            - RampThresholdProtocol (only inside Main Protocol)
            - Vecstim
            - Netstim

        extra_recs_vars (str): names of the extra recordings hoc variables
            Have the form ', var1, var2, ..., var_n' to be added to the hoc variable declaration
        extra_recs (str): hoc script to declare the extra recordings
    """

    def __init__(self, prot_definitions, mtype, add_synapses, apical_point_isec):
        """Get stimuli in hoc to put in createsimulation.hoc.

        Args:
            prot_definitions (dict): dictionary defining the protocols.
                Should have the structure of the protocol file in example/sscx/config/protocols
            mtype (str): mtype of the cell. prefix to use in the output files names
            add_synapses (bool): whether to add synapses to the cell
            apical_point_isec (int): section index of the apical point
                Set to -1 if there is no apical point
        """
        self.apical_point_isec = apical_point_isec
        self.n_stims = 0
        self.max_steps = 0
        self.reset_step_stimuli = ""
        self.init_step_stimuli = ""
        self.stims_hoc = ""
        self.extra_recs_vars = ""
        self.extra_recs = ""
        for prot_name, prot in prot_definitions.items():
            if "extra_recordings" in prot:
                self.add_extra_recs(prot["extra_recordings"])

            # reset stimuli and synapses before every protocol run
            self.stims_hoc += """
                reset_cell()
            """

            if "type" in prot and (
                prot["type"] == "StepProtocol"
                or prot["type"] == "StepThresholdProtocol"
            ):
                self.n_stims += 1
                step_hoc = self.get_step_hoc(prot)
                self.stims_hoc += step_hoc

                self.stims_hoc += """
                    simulate()
                """
                self.stims_hoc += self.add_save_recordings_hoc(mtype, prot_name, prot)

            elif "type" in prot and (
                prot["type"] == "RampProtocol"
                or prot["type"] == "RampThresholdProtocol"
            ):
                self.n_stims += 1
                ramp_hoc = self.get_ramp_hoc(prot)
                self.stims_hoc += ramp_hoc

                self.stims_hoc += """
                    simulate()
                """
                self.stims_hoc += self.add_save_recordings_hoc(mtype, prot_name, prot)

            elif "type" in prot and prot["type"] == "Vecstim" and add_synapses:
                self.n_stims += 1
                vecstim_hoc = self.get_vecstim_hoc(prot)
                self.stims_hoc += vecstim_hoc

                self.stims_hoc += """
                    simulate()
                """
                self.stims_hoc += self.add_save_recordings_hoc(mtype, prot_name, prot)

            elif "type" in prot and prot["type"] == "Netstim" and add_synapses:
                self.n_stims += 1
                netstim_hoc = self.get_netstim_hoc(prot)
                self.stims_hoc += netstim_hoc

                self.stims_hoc += """
                    simulate()
                """
                self.stims_hoc += self.add_save_recordings_hoc(mtype, prot_name, prot)

        self.get_reset_step()
        self.get_init_step()

    def add_extra_recs(self, extra_recs):
        """Add extra recordings to the recordings settings.

        Args:
            extra_recs (list): list of dicts defining the extra recordings
        """
        for extra_rec in extra_recs:
            name = extra_rec["name"]
            seclist_name = extra_rec["seclist_name"]
            var = extra_rec["var"]

            if name not in self.extra_recs_vars.split(", "):
                self.extra_recs_vars += f", {name}"
                if extra_rec["type"] == "nrnseclistcomp":
                    sec_index = extra_rec["sec_index"]
                    comp_x = extra_rec["comp_x"]

                    self.extra_recs += f"""
                        {name} = new Vector()
                        cell.{seclist_name}[{sec_index}] {name}.record(&{var}({comp_x}), 0.1)
                    """
                elif extra_rec["type"] == "somadistance":
                    somadistance = extra_rec["somadistance"]

                    self.extra_recs += f"""
                        {name} = new Vector()
                        secref = find_isec_at_soma_distance(cell.{seclist_name}, {somadistance})
                        comp_x = find_comp_x_at_soma_distance(secref, {somadistance})

                        secref.sec {name}.record(&{var}(comp_x), 0.1)
                    """

                elif extra_rec["type"] == "somadistanceapic":
                    somadistance = extra_rec["somadistance"]

                    self.extra_recs += f"""
                        {name} = new Vector()
                        apical_branch = get_apical_branch({self.apical_point_isec})
                        secref = find_isec_at_soma_distance(apical_branch, {somadistance})
                        comp_x = find_comp_x_at_soma_distance(secref, {somadistance})

                        secref.sec {name}.record(&{var}(comp_x), 0.1)
                    """

    @staticmethod
    def add_save_recordings_hoc(mtype, prot_name, prot):
        """Add this to the hoc file to save the recordings.

        Args:
            mtype (str): mtype of the cell. prefix to use in the output files names
            prot_name (str): name of the protocol. used in the output files names
            prot (dict): dictionary defining the protocol

        Returns:
            str: hoc scipt to save the recordings of given protocol.
        """
        save_recs = f"""
            sprint(fpath.s, "hoc_recordings/{mtype}.{prot_name}.soma.v.dat")
            timevoltage = new Matrix(time.size(), 2)
            timevoltage.setcol(0, time)
            timevoltage.setcol(1, voltage)
            write_output_file(fpath, timevoltage)
        """
        if "extra_recordings" in prot:
            for extra_rec in prot["extra_recordings"]:
                var = extra_rec["var"]
                name = extra_rec["name"]
                save_recs += f"""
                    sprint(fpath.s, "hoc_recordings/{mtype}.{prot_name}.{name}.{var}.dat")
                    timevoltage = new Matrix(time.size(), 2)
                    timevoltage.setcol(0, time)
                    timevoltage.setcol(1, {name})
                    write_output_file(fpath, timevoltage)
                """

        return save_recs

    def get_step_hoc(self, prot):
        """Get step stimuli in hoc format from step protocol dict.

        Args:
            prot (dict): dictionary defining the step protocol

        Returns:
            str: hoc script declaring one or more step stimuli
        """
        step_hoc = ""

        step_definitions = prot["stimuli"]["step"]
        if isinstance(step_definitions, dict):
            step_definitions = [step_definitions]
        for i, step in enumerate(step_definitions):
            self.max_steps = max(self.max_steps, i + 1)

            if step["amp"] is None:
                amp = f"{step['thresh_perc'] / 100.} * threshold_current"
            else:
                amp = step["amp"]

            step_hoc += f"""
                step_stimulus_{i} = new IClamp(0.5)
                step_stimulus_{i}.dur = {step["duration"]}
                step_stimulus_{i}.del = {step["delay"]}
                step_stimulus_{i}.amp = {amp}

                cell.soma step_stimulus_{i}
            """

        step_hoc += f"tstop={step_definitions[0]['totduration']}"

        if "holding" in prot["stimuli"]:
            hold = prot["stimuli"]["holding"]

            if hold["amp"] is None:
                amp = "holding_current"
            else:
                amp = hold["amp"]

            step_hoc += f"""
                holding_stimulus = new IClamp(0.5)
                holding_stimulus.dur = {hold["duration"]}
                holding_stimulus.del = {hold["delay"]}
                holding_stimulus.amp = {amp}

                cell.soma holding_stimulus
            """
        elif prot["type"] == "StepThresholdProtocol":
            step_hoc += f"""
                holding_stimulus = new IClamp(0.5)
                holding_stimulus.dur = {step_definitions[0]['totduration']}
                holding_stimulus.del = 0.0
                holding_stimulus.amp = holding_current

                cell.soma holding_stimulus
            """

        return step_hoc

    @staticmethod
    def get_ramp_hoc(prot):
        """Get ramp stimuli in hoc format from ramp protocol dict.

        Args:
            prot (dict): dictionary defining the ramp protocol

        Returns:
            str: hoc script declaring the ramp stimulus
        """
        ramp_hoc = ""

        ramp_definition = prot["stimuli"]["ramp"]
        # create time and amplitude of stimulus vectors

        if ramp_definition["ramp_amplitude_start"] is None:
            amp_start = (
                f"{ramp_definition['thresh_perc_start'] / 100.} * threshold_current"
            )
        else:
            amp_start = ramp_definition["ramp_amplitude_start"]
        if ramp_definition["ramp_amplitude_end"] is None:
            amp_end = f"{ramp_definition['thresh_perc_end'] / 100.} * threshold_current"
        else:
            amp_end = ramp_definition["ramp_amplitude_end"]

        ramp_hoc += """
            ramp_times = new Vector()
            ramp_amps = new Vector()

            ramp_times.append(0.0)
            ramp_amps.append(0.0)

            ramp_times.append({delay})
            ramp_amps.append(0.0)

            ramp_times.append({delay})
            ramp_amps.append({amplitude_start})

            ramp_times.append({delay} + {duration})
            ramp_amps.append({amplitude_end})

            ramp_times.append({delay} + {duration})
            ramp_amps.append(0.0)

            ramp_times.append({total_duration})
            ramp_amps.append(0.0)
        """.format(
            delay=ramp_definition["ramp_delay"],
            amplitude_start=amp_start,
            duration=ramp_definition["ramp_duration"],
            amplitude_end=amp_end,
            total_duration=ramp_definition["totduration"],
        )
        ramp_hoc += f"""
            ramp_stimulus = new IClamp(0.5)
            ramp_stimulus.dur = {ramp_definition["totduration"]}

            ramp_amps.play(&ramp_stimulus.amp, ramp_times, 1)

            cell.soma ramp_stimulus
        """

        ramp_hoc += f"tstop={ramp_definition['totduration']}"

        if "holding" in prot["stimuli"]:
            hold = prot["stimuli"]["holding"]

            if hold["amp"] is None:
                amp = "holding_current"
            else:
                amp = hold["amp"]

            ramp_hoc += f"""
                holding_stimulus = new IClamp(0.5)
                holding_stimulus.dur = {hold["duration"]}
                holding_stimulus.del = {hold["delay"]}
                holding_stimulus.amp = {amp}

                cell.soma holding_stimulus
            """
        elif prot["type"] == "RampThresholdProtocol":
            ramp_hoc += f"""
                holding_stimulus = new IClamp(0.5)
                holding_stimulus.dur = {ramp_definition['totduration']}
                holding_stimulus.del = 0.0
                holding_stimulus.amp = holding_current

                cell.soma holding_stimulus
            """

        return ramp_hoc

    @staticmethod
    def get_vecstim_hoc(prot):
        """Get vecstim stimulus in hoc format from vecstim protocol dict.

        Args:
            prot (dict): dictionary defining the vecstim protocol

        Returns:
            str: hoc script declaring the vecstim stimuli
        """
        stim = prot["stimuli"]

        vecstim_hoc = f"tstop={stim['syn_stop']}\n"

        hoc_synapse_creation = (
            "cell.synapses.create_netcons "
            + "({mode},{t0},{tf},{itv},{n_spike},{noise},{seed})"
        )
        vecstim_hoc += hoc_synapse_creation.format(
            mode=0,
            t0=stim["syn_start"],
            tf=stim["syn_stop"],
            itv=0,
            n_spike=0,
            noise=0,
            seed=stim["syn_stim_seed"],
        )

        return vecstim_hoc

    @staticmethod
    def get_netstim_hoc(prot):
        """Get netstim stimulus in hoc format from netstim protocol dict.

        Args:
            prot (dict): dictionary defining the netstim protocol

        Returns:
            str: hoc script declaring the netstim simuli
        """
        stim = prot["stimuli"]

        netstim_hoc = f"tstop={stim['syn_stop']}\n"

        hoc_synapse_creation = (
            "cell.synapses.create_netcons"
            + "({mode},{t0},{tf},{itv},{n_spike},{noise},{seed})"
        )
        netstim_hoc += hoc_synapse_creation.format(
            mode=1,
            t0=stim["syn_start"],
            tf=stim["syn_stop"],
            itv=stim["syn_interval"],
            n_spike=stim["syn_nmb_of_spikes"],
            noise=stim["syn_noise"],
            seed=0,
        )

        return netstim_hoc

    def get_reset_step(self):
        """Hoc script reseting all step stimuli needed by all the step protocols."""
        for i in range(max(self.max_steps, 1)):
            self.reset_step_stimuli += """
                step_stimulus_{i} = new IClamp(0.5)
                step_stimulus_{i}.dur = 0.0
                step_stimulus_{i}.del = 0.0
                step_stimulus_{i}.amp = 0.0

                cell.soma step_stimulus_{i}
            """.format(
                i=i
            )

    def get_init_step(self):
        """Hoc script initiating all step stimuli needed by all the step protocols."""
        for i in range(max(self.max_steps, 1)):
            self.init_step_stimuli += f"""
                objref step_stimulus_{i}
            """


def create_run_hoc(template_path, main_protocol):
    """Returns a string containing run.hoc.

    Args:
        template_path (str): path to the template to fill in
        main_protocol (bool): whether the Main Protocol is used or not

    Returns:
        str: hoc script to run the simulation
    """
    # load template
    with open(template_path, "r", encoding="utf-8") as template_file:
        template = template_file.read()
        template = jinja2.Template(template)

    # edit template
    return template.render(
        main_protocol=main_protocol,
    )


def create_synapse_hoc(
    syn_mech_args,
    syn_hoc_dir,
    template_path,
    gid,
    dt,
    synapses_template_name="hoc_synapses",
):
    """Returns a string containing the synapse hoc.

    Args:
        syn_mech_args (SynMechArgs): synapse-related configuration
        syn_hoc_dir (str): path to directory containing synapse-related data
        template_path (str): path to the template to fill in
        gid (int): cell ID
        dt (float): timestep (ms)
        synapses_template_name (str): template name of the synapse class

    Returns:
        str: hoc script with the synapse class template
    """
    # load template
    with open(template_path, "r", encoding="utf-8") as template_file:
        template = template_file.read()
        template = jinja2.Template(template)

    # edit template
    return template.render(
        TEMPLATENAME=synapses_template_name,
        GID=gid,
        SEED=syn_mech_args.seed,
        rng_settings_mode=syn_mech_args.rng_settings_mode,
        syn_dir=syn_hoc_dir,
        syn_conf_file=syn_mech_args.syn_conf_file,
        syn_data_file=syn_mech_args.syn_data_file,
        dt=dt,
    )


def create_hoc(
    mechs,
    parameters,
    ignored_globals=(),
    replace_axon=None,
    template_name="CCell",
    template_path="templates/cell_template.jinja2",
    disable_banner=None,
    add_synapses=False,
    synapses_template_name="hoc_synapses",
    syn_hoc_filename="synapses.hoc",
    syn_dir="synapses",
):
    """Return a string containing the hoc template.

    Args:
        mechs (list of bluepyopt.ephys.mechanisms.Mechanisms):
            All the mechs for the hoc template
        parameters (list of bluepyopt.Parameters): All the parameters in the hoc template
        ignored_globals (iterable str): HOC coded is added for each
            NrnGlobalParameter
            that exists, to test that it matches the values set in the parameters.
            This iterable contains parameter names that aren't checked
        replace_axon (str): String replacement for the 'replace_axon' command.
            Must include 'proc replace_axon(){ ... }
        template_name (str): name of cell class in hoc
        template_path (str): path to the jinja2 template
        disable_banner (bool): if not True: a banner is added to the hoc file
        add_synapses (bool): if True: synapses are loaded in the hoc
        synapses_template_name (str): synapse class name in hoc
        syn_hoc_filename (str): file name of synapse hoc file
        syn_dir (str): directory where the synapse data /files are

    Returns:
        str: hoc script describing the cell model
    """
    # pylint: disable=too-many-locals
    with open(template_path, "r", encoding="utf-8") as template_file:
        template = template_file.read()
        template = jinja2.Template(template)

    (
        global_params,
        section_params,
        range_params,
        _,
        location_order,
    ) = generate_parameters(parameters)
    range_params = range_exprs_to_hoc(range_params)
    channels, _ = generate_channels_by_location(mechs, location_order)

    ignored_global_params = {}
    for ignored_global in ignored_globals:
        if ignored_global in global_params:
            ignored_global_params[ignored_global] = global_params[ignored_global]
            del global_params[ignored_global]

    if not disable_banner:
        banner = f"Created by EModelRunner({__version__}) at {datetime.now()}"
    else:
        banner = None

    re_init_rng = generate_reinitrng(mechs)

    return template.render(
        template_name=template_name,
        banner=banner,
        channels=channels,
        section_params=section_params,
        range_params=range_params,
        global_params=global_params,
        re_init_rng=re_init_rng,
        replace_axon=replace_axon,
        ignored_global_params=ignored_global_params,
        add_synapses=add_synapses,
        synapses_template_name=synapses_template_name,
        syn_hoc_filename=syn_hoc_filename,
        syn_dir=syn_dir,
    )


def create_simul_hoc(
    template_path,
    add_synapses,
    hoc_paths,
    constants_args,
    protocol_definitions,
    apical_point_isec=-1,
):
    """Create createsimulation.hoc file.

    Args:
        template_path (str): path to the template to fill in
        add_synapses (bool): whether to add synapses to the cell
        hoc_paths (HocPaths): contains paths of the hoc files to be created
        constants_args (dict): contains data about the constants of the simulation
        protocol_definitions (dict): dictionary defining the protocols.
            Should have the structure of the protocol file in example/sscx/config/protocols
        apical_point_isec (int): section index of the apical point
            Set to -1 if there is no apical point

    Returns:
        str: hoc script to create the simulation
    """
    syn_dir = hoc_paths.syn_dir_for_hoc
    syn_hoc_file = hoc_paths.syn_hoc_filename
    cell_hoc_file = hoc_paths.cell_hoc_filename

    hoc_stim_creator = HocStimuliCreator(
        protocol_definitions,
        constants_args["mtype"],
        add_synapses,
        apical_point_isec,
    )

    # load template
    with open(template_path, "r", encoding="utf-8") as template_file:
        template = template_file.read()
        template = jinja2.Template(template)

    # edit template
    return template.render(
        cell_hoc_file=cell_hoc_file,
        add_synapses=add_synapses,
        syn_dir=syn_dir,
        syn_hoc_file=syn_hoc_file,
        celsius=constants_args["celsius"],
        v_init=constants_args["v_init"],
        dt=constants_args["dt"],
        template_name=constants_args["emodel"],
        gid=constants_args["gid"],
        morph_path=constants_args["morph_path"],
        run_simulation=hoc_stim_creator.stims_hoc,
        initiate_step_stimuli=hoc_stim_creator.init_step_stimuli,
        reset_step_stimuli=hoc_stim_creator.reset_step_stimuli,
        extra_recordings_vars=hoc_stim_creator.extra_recs_vars,
        extra_recordings=hoc_stim_creator.extra_recs,
    )


def create_main_protocol_hoc(
    template_path, protocol_definitions, rin_exp_voltage_base, mtype
):
    """Create main_protocol.hoc file.

    Args:
        template_path (str): path to the template to fill in
        protocol_definitions (dict): dictionary defining the protocols
        rin_exp_voltage_base (float): experimental value
            for the voltage_base feature for Rin protocol
        mtype (str): mtype of the cell. prefix to use in the output files names

    Returns:
        str: hoc script containing the main protocol functions
    """
    with open(template_path, "r", encoding="utf-8") as template_file:
        template = template_file.read()
        template = jinja2.Template(template)

    rmp_stim = protocol_definitions["RMP"]["stimuli"]
    rin_stim = protocol_definitions["Rin"]["stimuli"]
    thresh_stim = protocol_definitions["ThresholdDetection"]["step_template"]["stimuli"]

    # edit template
    return template.render(
        rmp_stimulus_dur=rmp_stim["step"]["duration"],
        rmp_stimulus_del=rmp_stim["step"]["delay"],
        rmp_stimulus_amp=rmp_stim["step"]["amp"],
        rmp_stimulus_totduration=rmp_stim["step"]["totduration"],
        rmp_output_path=f"hoc_recordings/{mtype}.RMP.soma.v.dat",
        rin_stimulus_dur=rin_stim["step"]["duration"],
        rin_stimulus_del=rin_stim["step"]["delay"],
        rin_stimulus_amp=rin_stim["step"]["amp"],
        rin_stimulus_totduration=rin_stim["step"]["totduration"],
        rin_holding_dur=rin_stim["holding"]["duration"],
        rin_holding_del=rin_stim["holding"]["delay"],
        holdi_precision=protocol_definitions["RinHoldcurrent"]["holdi_precision"],
        holdi_max_depth=protocol_definitions["RinHoldcurrent"]["holdi_max_depth"],
        voltagebase_exp_mean=rin_exp_voltage_base,
        rin_output_path=f"hoc_recordings/{mtype}.Rin.soma.v.dat",
        holding_current_output_path=f"hoc_recordings/{mtype}.bpo_holding_current.dat",
        thdetect_stimulus_del=thresh_stim["step"]["delay"],
        thdetect_stimulus_dur=thresh_stim["step"]["duration"],
        thdetect_stimulus_totduration=thresh_stim["step"]["totduration"],
        thdetect_holding_dur=thresh_stim["holding"]["duration"],
        thdetect_holding_del=thresh_stim["holding"]["delay"],
        threshold_current_output_path=f"hoc_recordings/{mtype}.bpo_threshold_current.dat",
    )



"""Run function to for synapse plasticity simulation."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import re

import numpy as np

from bluepyopt import ephys
from emodelrunner.create_cells import get_postcell
from emodelrunner.parsing_utilities import get_parser_args, set_verbosity
from emodelrunner.protocols.create_protocols import define_synapse_plasticity_protocols
from emodelrunner.load import get_release_params
from emodelrunner.load import get_syn_setup_params
from emodelrunner.load import load_config
from emodelrunner.output import write_synplas_output

# Configure logger
logger = logging.getLogger(__name__)


# taken from glusynapse.simulation.simulator
def _set_global_params(allparams, sim):
    """Set the global parameters of glusynapses.

    Args:
        allparams (dict): global glusynapse parameter to be set
        sim (bluepyopt.ephys.NrnSimulator): neuron simulator
    """
    logger.debug("Setting global parameters")
    for param_name, param_val in allparams.items():
        if re.match(".*_GluSynapse$", param_name):
            # Set parameter
            setattr(sim.neuron.h, param_name, param_val)
            logger.debug("\t%s = %f", param_name, getattr(sim.neuron.h, param_name))


def run(
    config_path,
    cvode_active=True,
    protocol_name="pulse",
    fixhp=True,
):
    """Run cell with pulse stimuli and pre-cell spike train.

    Args:
        config_path (str): path to config file
        cvode_active (bool): whether to use variable time step
        protocol_name (str): name of the protocol
        fixhp (bool): to uninsert SK_E2 for hyperpolarization in cell model
    """
    config = load_config(config_path=config_path)

    # load extra_params
    syn_setup_params = get_syn_setup_params(
        "synapses/syn_extra_params.json",
        "synapses/cpre_cpost.json",
        config.get("Paths", "synplas_fit_params_path"),
        config.getint("Cell", "gid"),
        config.getboolean("SynapsePlasticity", "invivo"),
    )

    # load cell
    cell = get_postcell(
        config=config,
        fixhp=fixhp,
        syn_setup_params=syn_setup_params,
    )

    sim = ephys.simulators.NrnSimulator(cvode_active=cvode_active)
    release_params = get_release_params(config)

    # set dynamic timestep tolerance
    sim.neuron.h.cvode.atolscale("v", 0.1)  # 0.01 for more precision

    # get pre_spike_train
    spike_train_path = config.get("Paths", "spiketrain_path")
    pre_spike_train = np.unique(np.loadtxt(spike_train_path, skiprows=1)[:, 0])

    # Set fitted model parameters
    if syn_setup_params["fit_params"]:
        _set_global_params(syn_setup_params["fit_params"], sim)

    # Enable in vivo mode (global)
    if config.getboolean("SynapsePlasticity", "invivo"):
        sim.neuron.h.cao_CR_GluSynapse = 1.2  # mM

    # protocols
    protocol = define_synapse_plasticity_protocols(
        cell,
        pre_spike_train,
        protocol_name,
        cvode_active,
        json.loads(config.get("SynapsePlasticity", "synrec")),
        config.getfloat("Protocol", "tstop"),
        config.getfloat("SynapsePlasticity", "fastforward"),
        config.get("Paths", "stimuli_path"),
    )

    # run
    logger.info("Python Recordings Running...")

    responses = protocol.run(
        cell_model=cell, param_values=release_params, sim=sim, isolate=False
    )

    # write responses
    output_path = config.get("Paths", "synplas_output_path")
    syn_prop_path = config.get("Paths", "syn_prop_path")
    write_synplas_output(responses, pre_spike_train, output_path, syn_prop_path)

    logger.info("Python Recordings Done.")
    logger.info("Responses written to %s", output_path)


if __name__ == "__main__":
    args = get_parser_args()
    set_verbosity(args.verbosity)

    run(config_path=args.config_path)



"""Protocol reading functions."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import json
from bluepyopt import ephys

from emodelrunner.protocols import sscx_protocols, thalamus_protocols
from emodelrunner.locations import SOMA_LOC
from emodelrunner.synapses.stimuli import (
    NrnNetStimStimulusCustom,
    NrnVecStimStimulusCustom,
)
from emodelrunner.protocols.protocols_func import (
    check_for_forbidden_protocol,
    get_recordings,
)


logger = logging.getLogger(__name__)


class ProtocolParser:
    """Parses the protocol json file."""

    def __init__(self) -> None:
        """Constructor.

        Attributes:
            protocols_dict (dict): protocols parsed into dict.
        """
        self.protocols_dict = {}

    @staticmethod
    def load_protocol_json(protocols_filepath):
        """Loads the protocol json file.

        Args:
            protocols_filepath (str or Path): path to the protocols file.

        Returns:
            dict: dict containing protocols json input.
        """
        with open(protocols_filepath, "r", encoding="utf-8") as protocol_file:
            protocol_definitions = json.load(protocol_file)

        if "__comment" in protocol_definitions:
            del protocol_definitions["__comment"]

        return protocol_definitions

    def _parse_step_and_ramp(
        self,
        protocol_definition,
        protocol_name,
        protocol_module,
        recordings,
        stochkv_det,
    ):
        """Parses the step and ramp protocols into self.protocols_dict."""
        if protocol_definition["type"] == "StepProtocol":
            self.protocols_dict[protocol_name] = read_step_protocol(
                protocol_name,
                protocol_module,
                protocol_definition,
                recordings,
                stochkv_det,
            )
        elif protocol_definition["type"] == "StepThresholdProtocol":
            self.protocols_dict[protocol_name] = read_step_threshold_protocol(
                protocol_name,
                protocol_module,
                protocol_definition,
                recordings,
                stochkv_det,
            )
        if protocol_module is sscx_protocols:
            if protocol_definition["type"] == "RampThresholdProtocol":
                self.protocols_dict[protocol_name] = read_ramp_threshold_protocol(
                    protocol_name,
                    protocol_definition,
                    recordings,
                )
            elif protocol_definition["type"] == "RampProtocol":
                self.protocols_dict[protocol_name] = read_ramp_protocol(
                    protocol_name, protocol_definition, recordings
                )

    def _parse_sscx_threshold_detection(self, protocol_definition, recordings, prefix):
        """Parses the sscx threshold detection protocol into self.protocols_dict."""
        if protocol_definition["type"] == "RatSSCxThresholdDetectionProtocol":
            self.protocols_dict["ThresholdDetection"] = (
                sscx_protocols.RatSSCxThresholdDetectionProtocol(
                    "IDRest",
                    step_protocol_template=read_step_protocol(
                        "Threshold",
                        sscx_protocols,
                        protocol_definition["step_template"],
                        recordings,
                    ),
                    prefix=prefix,
                )
            )

    def _parse_thalamus_threshold_detection(
        self, protocol_definition, protocol_name, recordings, prefix
    ):
        """Parses the thalamus threshold detection protocol into self.protocols_dict."""
        if protocol_definition["type"] == "RatSSCxThresholdDetectionProtocol":
            if protocol_name == "ThresholdDetection_dep":
                self.protocols_dict["ThresholdDetection_dep"] = (
                    thalamus_protocols.RatSSCxThresholdDetectionProtocol(
                        "ThresholdDetection_dep",
                        step_protocol_template=read_step_protocol(
                            "ThresholdDetection_dep",
                            thalamus_protocols,
                            protocol_definition["step_template"],
                            recordings,
                        ),
                        prefix=prefix,
                    )
                )
            elif protocol_name == "ThresholdDetection_hyp":
                self.protocols_dict["ThresholdDetection_hyp"] = (
                    thalamus_protocols.RatSSCxThresholdDetectionProtocol(
                        "ThresholdDetection_hyp",
                        step_protocol_template=read_step_protocol(
                            "ThresholdDetection_hyp",
                            thalamus_protocols,
                            protocol_definition["step_template"],
                            recordings,
                        ),
                        prefix=prefix,
                    )
                )

    def _parse_vecstim_netstim(
        self, protocol_definition, protocol_name, recordings, syn_locs
    ):
        """Parses the vecstim netstim protocols into self.protocols_dict."""
        if protocol_definition["type"] == "Vecstim":
            self.protocols_dict[protocol_name] = read_vecstim_protocol(
                protocol_name, protocol_definition, recordings, syn_locs
            )
        elif protocol_definition["type"] == "Netstim":
            self.protocols_dict[protocol_name] = read_netstim_protocol(
                protocol_name, protocol_definition, recordings, syn_locs
            )

    def _parse_sscx_main(self, protocol_definitions, prefix):
        """Parses the main sscx protocol into self.protocols_dict."""
        self.protocols_dict["RinHoldcurrent"] = (
            sscx_protocols.RatSSCxRinHoldcurrentProtocol(
                "RinHoldCurrent",
                rin_protocol_template=self.protocols_dict["Rin"],
                holdi_precision=protocol_definitions["RinHoldcurrent"][
                    "holdi_precision"
                ],
                holdi_max_depth=protocol_definitions["RinHoldcurrent"][
                    "holdi_max_depth"
                ],
                prefix=prefix,
            )
        )

        other_protocols = []

        for protocol_name in protocol_definitions["Main"]["other_protocols"]:
            if protocol_name in self.protocols_dict:
                other_protocols.append(self.protocols_dict[protocol_name])

        pre_protocols = []

        if "pre_protocols" in protocol_definitions["Main"]:
            for protocol_name in protocol_definitions["Main"]["pre_protocols"]:
                pre_protocols.append(self.protocols_dict[protocol_name])

        self.protocols_dict["Main"] = sscx_protocols.RatSSCxMainProtocol(
            "Main",
            rmp_protocol=self.protocols_dict["RMP"],
            rinhold_protocol=self.protocols_dict["RinHoldcurrent"],
            thdetect_protocol=self.protocols_dict["ThresholdDetection"],
            other_protocols=other_protocols,
            pre_protocols=pre_protocols,
        )

    def _parse_thalamus_main(self, protocol_definitions, prefix):
        """Parses the main thalamus protocol into self.protocols_dict."""
        try:  # Only low-threshold bursting cells have thin protocol
            self.protocols_dict["RinHoldcurrent_dep"] = (
                thalamus_protocols.RatSSCxRinHoldcurrentProtocol(
                    "RinHoldcurrent_dep",
                    rin_protocol_template=self.protocols_dict["Rin_dep"],
                    holdi_estimate_multiplier=protocol_definitions[
                        "RinHoldcurrent_dep"
                    ]["holdi_estimate_multiplier"],
                    holdi_precision=protocol_definitions["RinHoldcurrent_dep"][
                        "holdi_precision"
                    ],
                    holdi_max_depth=protocol_definitions["RinHoldcurrent_dep"][
                        "holdi_max_depth"
                    ],
                    prefix=prefix,
                )
            )
            rinhold_protocol_dep = self.protocols_dict["RinHoldcurrent_dep"]
            thdetect_protocol_dep = self.protocols_dict["ThresholdDetection_dep"]
        except KeyError:
            rinhold_protocol_dep = None
            thdetect_protocol_dep = None

        self.protocols_dict["RinHoldcurrent_hyp"] = (
            thalamus_protocols.RatSSCxRinHoldcurrentProtocol(
                "RinHoldcurrent_hyp",
                rin_protocol_template=self.protocols_dict["Rin_hyp"],
                holdi_estimate_multiplier=protocol_definitions["RinHoldcurrent_hyp"][
                    "holdi_estimate_multiplier"
                ],
                holdi_precision=protocol_definitions["RinHoldcurrent_hyp"][
                    "holdi_precision"
                ],
                holdi_max_depth=protocol_definitions["RinHoldcurrent_hyp"][
                    "holdi_max_depth"
                ],
                prefix=prefix,
            )
        )

        other_protocols = [
            self.protocols_dict[protocol_name]
            for protocol_name in protocol_definitions["Main"]["other_protocols"]
        ]

        pre_protocols = []
        if "pre_protocols" in protocol_definitions["Main"]:
            for protocol_name in protocol_definitions["Main"]["pre_protocols"]:
                pre_protocols.append(self.protocols_dict[protocol_name])

        self.protocols_dict["Main"] = thalamus_protocols.RatSSCxMainProtocol(
            "Main",
            rmp_protocol=self.protocols_dict["RMP"],
            rinhold_protocol_dep=rinhold_protocol_dep,
            rinhold_protocol_hyp=self.protocols_dict["RinHoldcurrent_hyp"],
            thdetect_protocol_dep=thdetect_protocol_dep,
            thdetect_protocol_hyp=self.protocols_dict["ThresholdDetection_hyp"],
            other_protocols=other_protocols,
            pre_protocols=pre_protocols,
        )

    def parse_sscx_protocols(
        self,
        protocols_filepath,
        stochkv_det=None,
        prefix="",
        apical_point_isec=-1,
        syn_locs=None,
    ):
        """Parses the SSCX protocols from the json file input.

        Args:
            protocols_filename (str): path to the protocols file
            stochkv_det (bool): set if stochastic or deterministic
            prefix (str): prefix used in naming responses, features, recordings, etc.
            apical_point_isec (int): apical point section index
                Should be given if there is "somadistanceapic" in "type"
                of at least one of the extra recordings
            syn_locs (list of ephys.locations.NrnPointProcessLocation):
                locations of the synapses (if any, else None)

        Returns:
            dict containing the protocols
        """
        protocol_definitions = self.load_protocol_json(protocols_filepath)

        for protocol_name, protocol_definition in protocol_definitions.items():
            if protocol_name not in ["Main", "RinHoldcurrent"]:
                recordings = get_recordings(
                    protocol_name, protocol_definition, prefix, apical_point_isec
                )

                if "type" in protocol_definition:
                    # add protocol to protocol dict
                    self._parse_step_and_ramp(
                        protocol_definition,
                        protocol_name,
                        sscx_protocols,
                        recordings,
                        stochkv_det,
                    )
                    self._parse_sscx_threshold_detection(
                        protocol_definition, recordings, prefix
                    )
                    self._parse_vecstim_netstim(
                        protocol_definition, protocol_name, recordings, syn_locs
                    )

                else:
                    stimuli = [
                        ephys.stimuli.NrnSquarePulse(
                            step_amplitude=stimulus_definition["amp"],
                            step_delay=stimulus_definition["delay"],
                            step_duration=stimulus_definition["duration"],
                            location=SOMA_LOC,
                            total_duration=stimulus_definition["totduration"],
                        )
                        for stimulus_definition in protocol_definition["stimuli"]
                    ]
                    self.protocols_dict[protocol_name] = ephys.protocols.SweepProtocol(
                        name=protocol_name, stimuli=stimuli, recordings=recordings
                    )

        if "Main" in protocol_definitions.keys():
            self._parse_sscx_main(protocol_definitions, prefix)
        else:
            check_for_forbidden_protocol(self.protocols_dict)

        return self.protocols_dict

    def parse_thalamus_protocols(
        self,
        protocols_filepath,
        stochkv_det=None,
        prefix="",
    ):
        """Parses the Thalamus protocols from the json file input.

        Args:
            protocols_filename (str): path to the protocols file
            stochkv_det (bool): set if stochastic or deterministic
            prefix (str): prefix used in naming responses, features, recordings, etc.

        Returns:
            dict containing the protocols
        """
        protocol_definitions = self.load_protocol_json(protocols_filepath)

        for protocol_name, protocol_definition in protocol_definitions.items():
            if protocol_name not in [
                "Main",
                "RinHoldcurrent_dep",
                "RinHoldcurrent_hyp",
            ]:
                # By default include somatic recording
                somav_recording = ephys.recordings.CompRecording(
                    name=f"{prefix}.{protocol_name}.soma.v",
                    location=SOMA_LOC,
                    variable="v",
                )

                recordings = [somav_recording]

                if "type" in protocol_definition:
                    # add protocol to protocol dict
                    self._parse_step_and_ramp(
                        protocol_definition,
                        protocol_name,
                        thalamus_protocols,
                        recordings,
                        stochkv_det,
                    )
                    self._parse_thalamus_threshold_detection(
                        protocol_definition, protocol_name, recordings, prefix
                    )

                else:
                    stimuli = [
                        ephys.stimuli.NrnSquarePulse(
                            step_amplitude=stimulus_definition["amp"],
                            step_delay=stimulus_definition["delay"],
                            step_duration=stimulus_definition["duration"],
                            location=SOMA_LOC,
                            total_duration=stimulus_definition["totduration"],
                        )
                        for stimulus_definition in protocol_definition["stimuli"]
                    ]
                    self.protocols_dict[protocol_name] = ephys.protocols.SweepProtocol(
                        name=protocol_name, stimuli=stimuli, recordings=recordings
                    )

        if "Main" in protocol_definitions.keys():
            self._parse_thalamus_main(protocol_definitions, prefix)
        else:
            check_for_forbidden_protocol(self.protocols_dict)

        return self.protocols_dict


def read_ramp_threshold_protocol(protocol_name, protocol_definition, recordings):
    """Read ramp threshold protocol from definition.

    Args:
        protocol_name (str): name of the protocol
        protocol_definition (dict): contains the protocol configuration data
        recordings (bluepyopt.ephys.recordings.CompRecording):
            recordings to use with this protocol

    Returns:
        RampThresholdProtocol: Ramp Protocol depending on cell's threshold current
    """
    ramp_definition = protocol_definition["stimuli"]["ramp"]
    ramp_stimulus = ephys.stimuli.NrnRampPulse(
        ramp_delay=ramp_definition["ramp_delay"],
        ramp_duration=ramp_definition["ramp_duration"],
        location=SOMA_LOC,
        total_duration=ramp_definition["totduration"],
    )

    holding_stimulus = ephys.stimuli.NrnSquarePulse(
        step_delay=0.0,
        step_duration=ramp_definition["totduration"],
        location=SOMA_LOC,
        total_duration=ramp_definition["totduration"],
    )

    return sscx_protocols.RampThresholdProtocol(
        name=protocol_name,
        ramp_stimulus=ramp_stimulus,
        holding_stimulus=holding_stimulus,
        thresh_perc_start=ramp_definition["thresh_perc_start"],
        thresh_perc_end=ramp_definition["thresh_perc_end"],
        recordings=recordings,
    )


def read_ramp_protocol(protocol_name, protocol_definition, recordings):
    """Read ramp protocol from definition.

    Args:
        protocol_name (str): name of the protocol
        protocol_definition (dict): contains the protocol configuration data
        recordings (bluepyopt.ephys.recordings.CompRecording):
            recordings to use with this protocol

    Returns:
        RampProtocol: Ramp Protocol
    """
    ramp_definition = protocol_definition["stimuli"]["ramp"]
    ramp_stimulus = ephys.stimuli.NrnRampPulse(
        ramp_amplitude_start=ramp_definition["ramp_amplitude_start"],
        ramp_amplitude_end=ramp_definition["ramp_amplitude_end"],
        ramp_delay=ramp_definition["ramp_delay"],
        ramp_duration=ramp_definition["ramp_duration"],
        location=SOMA_LOC,
        total_duration=ramp_definition["totduration"],
    )

    if "holding" in protocol_definition["stimuli"]:
        holding_definition = protocol_definition["stimuli"]["holding"]
        holding_stimulus = ephys.stimuli.NrnSquarePulse(
            step_amplitude=holding_definition["amp"],
            step_delay=holding_definition["delay"],
            step_duration=holding_definition["duration"],
            location=SOMA_LOC,
            total_duration=holding_definition["totduration"],
        )
    else:
        holding_stimulus = None

    return sscx_protocols.RampProtocol(
        name=protocol_name,
        ramp_stimulus=ramp_stimulus,
        holding_stimulus=holding_stimulus,
        recordings=recordings,
    )


def read_step_protocol(
    protocol_name, protocol_module, protocol_definition, recordings, stochkv_det=None
):
    """Read step protocol from definition.

    Args:
        protocol_name (str): name of the protocol
        protocol_module (module): module that contains the protocol
        protocol_definition (dict): contains the protocol configuration data
        recordings (bluepyopt.ephys.recordings.CompRecording):
            recordings to use with this protocol
        protocol_cls (module): module that contains the protocol
        stochkv_det (bool): set if stochastic or deterministic

    Returns:
        sscx_protocols.StepProtocol or thalamus_protocols.StepProtocolCustom: Step Protocol
    """
    # pylint: disable=undefined-loop-variable
    step_definitions = protocol_definition["stimuli"]["step"]
    if isinstance(step_definitions, dict):
        step_definitions = [step_definitions]

    step_stimuli = []
    for step_definition in step_definitions:
        step_stim = ephys.stimuli.NrnSquarePulse(
            step_amplitude=step_definition["amp"],
            step_delay=step_definition["delay"],
            step_duration=step_definition["duration"],
            location=SOMA_LOC,
            total_duration=step_definition["totduration"],
        )
        step_stimuli.append(step_stim)

    if "holding" in protocol_definition["stimuli"]:
        holding_definition = protocol_definition["stimuli"]["holding"]
        holding_stimulus = ephys.stimuli.NrnSquarePulse(
            step_amplitude=holding_definition["amp"],
            step_delay=holding_definition["delay"],
            step_duration=holding_definition["duration"],
            location=SOMA_LOC,
            total_duration=holding_definition["totduration"],
        )
    else:
        holding_stimulus = None

    if stochkv_det is None:
        stochkv_det = (
            step_definition["stochkv_det"] if "stochkv_det" in step_definition else None
        )

    if protocol_module is thalamus_protocols:
        return protocol_module.StepProtocolCustom(
            name=protocol_name,
            step_stimulus=step_stimuli[0],
            holding_stimulus=holding_stimulus,
            recordings=recordings,
            stochkv_det=stochkv_det,
        )
    elif protocol_module is sscx_protocols:
        return protocol_module.StepProtocol(
            name=protocol_name,
            step_stimuli=step_stimuli,
            holding_stimulus=holding_stimulus,
            recordings=recordings,
            stochkv_det=stochkv_det,
        )
    else:
        raise ValueError(f"unsupported protocol module: {protocol_module}")


def read_step_threshold_protocol(
    protocol_name, protocol_module, protocol_definition, recordings, stochkv_det=None
):
    """Read step threshold protocol from definition.

    Args:
        protocol_name (str): name of the protocol
        protocol_module (module): module that contains the protocol
        protocol_definition (dict): contains the protocol configuration data
        recordings (bluepyopt.ephys.recordings.CompRecording): recordings to use with this protocol
        stochkv_det (bool): set if stochastic or deterministic

    Returns:
        sscx_protocols.StepThresholdProtocol
         or thalamus_protocols.StepThresholdProtocol: StepProtocol with cell's threshold current
    """
    # pylint: disable=undefined-loop-variable
    step_definitions = protocol_definition["stimuli"]["step"]
    if isinstance(step_definitions, dict):
        step_definitions = [step_definitions]

    step_stimuli = []
    for step_definition in step_definitions:
        step_stim = ephys.stimuli.NrnSquarePulse(
            step_delay=step_definition["delay"],
            step_duration=step_definition["duration"],
            location=SOMA_LOC,
            total_duration=step_definition["totduration"],
        )
        step_stimuli.append(step_stim)

    holding_stimulus = ephys.stimuli.NrnSquarePulse(
        step_delay=0.0,
        step_duration=step_definition["totduration"],
        location=SOMA_LOC,
        total_duration=step_definition["totduration"],
    )

    if stochkv_det is None:
        stochkv_det = (
            step_definition["stochkv_det"] if "stochkv_det" in step_definition else None
        )

    if protocol_module is thalamus_protocols:
        return protocol_module.StepThresholdProtocol(
            name=protocol_name,
            step_stimulus=step_stimuli[0],
            holding_stimulus=holding_stimulus,
            thresh_perc=step_definition["thresh_perc"],
            recordings=recordings,
            stochkv_det=stochkv_det,
        )
    elif protocol_module is sscx_protocols:
        return protocol_module.StepThresholdProtocol(
            name=protocol_name,
            step_stimuli=step_stimuli,
            holding_stimulus=holding_stimulus,
            thresh_perc=step_definition["thresh_perc"],
            recordings=recordings,
            stochkv_det=stochkv_det,
        )
    else:
        raise ValueError(f"unsupported protocol module: {protocol_module}")


def read_vecstim_protocol(protocol_name, protocol_definition, recordings, syn_locs):
    """Read Vecstim protocol from definitions.

    Args:
        protocol_name (str): name of the protocol
        protocol_definition (dict): dict containing the protocol data
        recordings (bluepyopt.ephys.recordings.CompRecording):
            recordings to use with this protocol
        syn_locs (list of ephys.locations.NrnPointProcessLocation):
            locations of the synapses

    Returns:
        emodelrunner.protocols.SweepProtocolCustom:
            a protocol containing Vecstim stimulus activating synapses
    """
    stim_definition = protocol_definition["stimuli"]
    if stim_definition["vecstim_random"] not in [
        "python",
        "neuron",
    ]:
        logger.warning(
            "vecstim random not set to 'python' nor to 'neuron' in config file."
            "vecstim random will be re-set to 'python'."
        )
        stim_definition["vecstim_random"] = "python"

    stim = NrnVecStimStimulusCustom(
        syn_locs,
        stim_definition["syn_start"],
        stim_definition["syn_stop"],
        stim_definition["syn_stim_seed"],
        stim_definition["vecstim_random"],
    )

    return sscx_protocols.SweepProtocolCustom(protocol_name, [stim], recordings)


def read_netstim_protocol(protocol_name, protocol_definition, recordings, syn_locs):
    """Read Netstim protocol from definitions.

    Args:
        protocol_name (str): name of the protocol
        protocol_definition (dict): dict containing the protocol data
        recordings (bluepyopt.ephys.recordings.CompRecording):
            recordings to use with this protocol
        syn_locs (list of ephys.locations.NrnPointProcessLocation):
            locations of the synapses

    Returns:
        emodelrunner.protocols.SweepProtocolCustom:
            a protocol containing Netstim stimulus activating synapses
    """
    stim_definition = protocol_definition["stimuli"]

    stim = NrnNetStimStimulusCustom(
        syn_locs,
        stim_definition["syn_stop"],
        stim_definition["syn_nmb_of_spikes"],
        stim_definition["syn_interval"],
        stim_definition["syn_start"],
        stim_definition["syn_noise"],
    )

    return sscx_protocols.SweepProtocolCustom(protocol_name, [stim], recordings)



"""GluSynapse class."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from emodelrunner.synapses.synapse import SynapseMixin


class GluSynapseCustom(SynapseMixin):
    """Attach a synapse to the simulation.

    Attributes:
        seed (int): random number generator seed number
        rng_settins_mode (str) : mode of the random number generator
            Can be "Random123" or "Compatibility"
        section (neuron section): cell location where the synapse is attached to
        hsynapse (neuron GluSynapse): Glusynapse instantion in simulator
        delay (float): synapse delay
        weight (float): synapse weight
        pre_mtype (int): ID (but not gid) of the presynaptic cell
        start (int/None): force synapse to start firing at given value when using NetStim
        interval (int/None): force synapse to fire at given interval when using NetStim
        number (int/None): force synapse to fire N times when using NetStim
        noise (int/None): force synapse to have given noise when using NetStim
    """

    def __init__(
        self,
        sim,
        icell,
        synapse,
        section,
        seed,
        rng_settings_mode,
        synconf_dict,
        start=None,
        interval=None,
        number=None,
        noise=None,
    ):
        """Constructor.

        Args:
            sim (NrnSimulator): simulator
            icell (Hoc Cell): cell to which attach the synapse
            synapse (dict): synapse data
            section (neuron section): cell location where the synapse is attached to
            seed (int): random number generator seed number
            rng_settings_mode (str): mode of the random number generator
                Can be "Random123" or "Compatibility"
            synconf_dict (dict): synapse configuration
            start (int/None): force synapse to start firing at given value when using NetStim
            interval (int/None): force synapse to fire at given interval when using NetStim
            number (int/None): force synapse to fire N times when using NetStim
            noise (int/None): force synapse to have given noise when using NetStim
        """
        # pylint: disable=too-many-arguments
        self.seed = seed
        self.rng_settings_mode = rng_settings_mode
        self.section = section

        # the synapse is inhibitory
        if synapse["synapse_type"] < 100:
            raise NotImplementedError()
        # the synapse is excitatory
        self.hsynapse = sim.neuron.h.GluSynapse(synapse["seg_x"], sec=self.section)
        self.hsynapse.tau_d_AMPA = synapse["tau_d"]

        self.hsynapse.Use0_TM = abs(synapse["use"])
        self.hsynapse.Dep_TM = abs(synapse["dep"])
        self.hsynapse.Fac_TM = abs(synapse["fac"])

        self.hsynapse.synapseID = synapse["sid"]

        self.hsynapse.Nrrp_TM = synapse["Nrrp"]

        # set random number generator
        self.set_random_nmb_generator(sim, icell, synapse["sid"])

        self.execute_synapse_configuration(synconf_dict, synapse["sid"], sim)

        self.delay = synapse["delay"]
        self.weight = synapse["weight"]

        self.pre_mtype = synapse["pre_mtype"]

        # netstim params if given
        self.start = start
        self.interval = interval
        self.number = number
        self.noise = noise

    # taken from glusynapseutils.simulation.simulator._runconnectedpair_process
    def set_local_params(self, fit_params, extra_params, c_pre=0.0, c_post=0.0):
        """Set local parameters of given synapse.

        Args:
            fit_params (dict): glusynapse parameters from fitting to get threshold values
            extra_params (dict): contains synapse location and synapse extra parameters
            c_pre (float): calcium peak during a single EPSP
            c_post (float): calcium peak during a single bAP
        """
        # Update basic synapse parameters
        for param in extra_params:
            if param == "loc":
                continue
            # Set parameter
            setattr(self.hsynapse, param, extra_params[param])

        # Update other parameters
        if fit_params is not None:
            if (
                all(key in fit_params for key in ["a00", "a01"])
                and extra_params["loc"] == "basal"
            ):
                # Set basal depression threshold
                self.hsynapse.theta_d_GB = (
                    fit_params["a00"] * c_pre + fit_params["a01"] * c_post
                )
            if (
                all(key in fit_params for key in ["a10", "a11"])
                and extra_params["loc"] == "basal"
            ):
                # Set basal potentiation threshold
                self.hsynapse.theta_p_GB = (
                    fit_params["a10"] * c_pre + fit_params["a11"] * c_post
                )
            if (
                all(key in fit_params for key in ["a20", "a21"])
                and extra_params["loc"] == "apical"
            ):
                # Set apical depression threshold
                self.hsynapse.theta_d_GB = (
                    fit_params["a20"] * c_pre + fit_params["a21"] * c_post
                )
            if (
                all(key in fit_params for key in ["a30", "a31"])
                and extra_params["loc"] == "apical"
            ):
                # Set apical potentiation threshold
                self.hsynapse.theta_p_GB = (
                    fit_params["a30"] * c_pre + fit_params["a31"] * c_post
                )

    # taken from glusynapse.simulation.simulator._runconnectedpair_process
    def setup_synapses(self, params):
        """Set local parameters of given synapse from params dict.

        Args:
            params (dict): contains

                - fit_params: glusynapse parameters from fitting to get threshold values
                - syn_extra_params: synapse location and synapse extra parameters
                - c_pre: calcium peak during a single EPSP
                - c_post: calcium peak during a single bAP
                - postgid: ID of the postsynaptic cell
                - invivo: whether to put synapse in 'in vivo' conditions
        """
        syn_id = int(self.hsynapse.synapseID)
        # Set local parameters
        key = str((params["postgid"], syn_id))
        self.set_local_params(
            params["fit_params"],
            params["syn_extra_params"][key],
            params["c_pre"][key],
            params["c_post"][key],
        )
        # Enable in vivo mode (synapse)
        if params["invivo"]:
            self.hsynapse.Use0_TM = 0.15 * self.hsynapse.Use0_TM
            self.hsynapse.Use_d_TM = 0.15 * self.hsynapse.Use_d_TM
            self.hsynapse.Use_p_TM = 0.15 * self.hsynapse.Use_p_TM



"""Parsing utilities functions."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging


def get_parser_args():
    """Get config_path and verbosity from argparse.

    Returns:
        argparse.Namespace: object containing the parsed arguments
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config_path",
        default=None,
        help="the path to the config file.",
    )
    parser.add_argument("-v", "--verbose", type=int, dest="verbosity", default=1)
    return parser.parse_args()


def set_verbosity(verbosity):
    """Set verbosity level.

    Args:
        verbosity (int): verbosity level. 0 for warning, 1 for info and 2 or more for debug
    """
    if verbosity > 2:
        verbosity = 2
    elif verbosity < 0:
        verbosity = 0

    logging.basicConfig(
        level=(logging.WARNING, logging.INFO, logging.DEBUG)[verbosity],
        handlers=[logging.StreamHandler()],
    )



"""Locations related functions and classes."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from bluepyopt import ephys

logger = logging.getLogger(__name__)


SOMA_LOC = ephys.locations.NrnSeclistCompLocation(
    name="soma", seclist_name="somatic", sec_index=0, comp_x=0.5
)


def multi_locations(sectionlist):
    """Define locations.

    Args:
        sectionlist (str): Name of the location(s) to return.
            Can be alldend, somadend, somaxon, allact, apical, basal, somatic, axonal

    Returns:
        list: locations corresponding to sectionlist
    """
    if sectionlist == "alldend":
        seclist_locs = [
            ephys.locations.NrnSeclistLocation("apical", seclist_name="apical"),
            ephys.locations.NrnSeclistLocation("basal", seclist_name="basal"),
        ]
    elif sectionlist == "somadend":
        seclist_locs = [
            ephys.locations.NrnSeclistLocation("apical", seclist_name="apical"),
            ephys.locations.NrnSeclistLocation("basal", seclist_name="basal"),
            ephys.locations.NrnSeclistLocation("somatic", seclist_name="somatic"),
        ]
    elif sectionlist == "somaxon":
        seclist_locs = [
            ephys.locations.NrnSeclistLocation("axonal", seclist_name="axonal"),
            ephys.locations.NrnSeclistLocation("somatic", seclist_name="somatic"),
        ]
    elif sectionlist == "allact":
        seclist_locs = [
            ephys.locations.NrnSeclistLocation("apical", seclist_name="apical"),
            ephys.locations.NrnSeclistLocation("basal", seclist_name="basal"),
            ephys.locations.NrnSeclistLocation("somatic", seclist_name="somatic"),
            ephys.locations.NrnSeclistLocation("axonal", seclist_name="axonal"),
        ]
    else:
        seclist_locs = [
            ephys.locations.NrnSeclistLocation(sectionlist, seclist_name=sectionlist)
        ]

    return seclist_locs



"""Functions to create recordings."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from bluepyopt import ephys

from emodelrunner.synapses.recordings import SynapseRecordingCustom


def get_pairsim_recordings(
    soma_loc, syn_locs, synrecs, presyn_prot_name, postsyn_prot_name
):
    """Return the precell and the postcell recordings for a pair simulation.

    Args:
        soma_loc (bluepyopt.ephys.locations.NrnSeclistCompLocation):
            location of the soma of the pre-synaptic cell
        syn_locs (list of bluepyopt.ephys.locations.NrnPointProcessLocation):
            location of synapses  of the post-synaptic cell
        synrecs (list of str): the extra synapse variables to record
        presyn_prot_name (str): presynaptic protocol name
        postsyn_prot_name (str): postsynaptic protocol name

    Returns:
        a tuple containing

        - list of recordings: presynaptic recordings
        - list of recordings: postsynaptic recordings
    """
    presyn_rec = ephys.recordings.CompRecording(
        name=presyn_prot_name, location=soma_loc, variable="v"
    )
    presyn_recs = [presyn_rec]
    postsyn_rec = ephys.recordings.CompRecording(
        name=postsyn_prot_name, location=soma_loc, variable="v"
    )
    postsyn_recs = [postsyn_rec]

    for syn_loc in syn_locs:
        for synrec in synrecs:
            postsyn_recs.append(
                SynapseRecordingCustom(name=synrec, location=syn_loc, variable=synrec)
            )

    return (presyn_recs, postsyn_recs)
